# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from re import match
from json import dumps
from ..decorators import requires_auth
from .comment import IssueComment, issue_comment_params
from .event import IssueEvent
from .label import Label
from .milestone import Milestone
from ..models import GitHubCore
from ..users import User
from uritemplate import URITemplate


class Issue(GitHubCore):

    """The :class:`Issue <Issue>` object. It structures and handles the data
    returned via the `Issues <http://developer.github.com/v3/issues>`_ section
    of the GitHub API.

    Two issue instances can be checked like so::

        i1 == i2
        i1 != i2

    And is equivalent to::

        i1.id == i2.id
        i1.id != i2.id

    """

    # The Accept header will likely be removable once the feature is out of
    # preview mode. See: https://git.io/vgXmB
    LOCKING_PREVIEW_HEADERS = {
        'Accept': 'application/vnd.github.the-key-preview+json'
    }

    def _update_attributes(self, issue):
        self._api = issue.get('url', '')
        #: :class:`User <github3.users.User>` representing the user the issue
        #: was assigned to.
        self.assignee = issue.get('assignee')
        if self.assignee:
            self.assignee = User(issue.get('assignee'), self)
        #: Body (description) of the issue.
        self.body = issue.get('body', '')
        #: HTML formatted body of the issue.
        self.body_html = issue.get('body_html', '')
        #: Plain text formatted body of the issue.
        self.body_text = issue.get('body_text', '')

        # If an issue is still open, this field will be None
        #: datetime object representing when the issue was closed.
        self.closed_at = self._strptime(issue.get('closed_at'))

        #: Number of comments on this issue.
        self.comments_count = issue.get('comments')
        #: Comments url (not a template)
        self.comments_url = issue.get('comments_url')
        #: datetime object representing when the issue was created.
        self.created_at = self._strptime(issue.get('created_at'))
        #: Events url (not a template)
        self.events_url = issue.get('events_url')
        #: URL to view the issue at GitHub.
        self.html_url = issue.get('html_url')
        #: Unique ID for the issue.
        self.id = issue.get('id')
        #: Returns the list of :class:`Label <github3.issues.label.Label>`\ s
        #: on this issue.
        self.original_labels = [
            Label(l, self) for l in issue.get('labels')
        ]
        labels_url = issue.get('labels_url')
        #: Labels URL Template. Expand with ``name``
        self.labels_urlt = URITemplate(labels_url) if labels_url else None
        #: Locked status
        self.locked = issue.get('locked')
        #: :class:`Milestone <github3.issues.milestone.Milestone>` this
        #: issue was assigned to.
        self.milestone = None
        if issue.get('milestone'):
            self.milestone = Milestone(issue.get('milestone'), self)
        #: Issue number (e.g. #15)
        self.number = issue.get('number')
        #: Dictionary URLs for the pull request (if they exist)
        self.pull_request_urls = issue.get('pull_request', {})
        m = match('https?://[\w\d\-\.\:]+/(\S+)/(\S+)/(?:issues|pull)/\d+',
                  self.html_url)
        #: Returns ('owner', 'repository') this issue was filed on.
        self.repository = m.groups()
        #: State of the issue, e.g., open, closed
        self.state = issue.get('state')
        #: Title of the issue.
        self.title = issue.get('title')
        #: datetime object representing the last time the issue was updated.
        self.updated_at = self._strptime(issue.get('updated_at'))
        #: :class:`User <github3.users.User>` who opened the issue.
        self.user = User(issue.get('user'), self)

        closed_by = issue.get('closed_by')
        #: :class:`User <github3.users.User>` who closed the issue.
        self.closed_by = User(closed_by, self) if closed_by else None

    def _repr(self):
        return '<Issue [{r[0]}/{r[1]} #{n}]>'.format(r=self.repository,
                                                     n=self.number)

    @requires_auth
    def add_labels(self, *args):
        """Add labels to this issue.

        :param str args: (required), names of the labels you wish to add
        :returns: list of :class:`Label`\ s
        """
        url = self._build_url('labels', base_url=self._api)
        json = self._json(self._post(url, data=args), 200)
        return [Label(l, self) for l in json] if json else []

    @requires_auth
    def assign(self, username):
        """Assigns user ``username`` to this issue. This is a short cut for
        ``issue.edit``.

        :param str username: username of the person to assign this issue to
        :returns: bool
        """
        if not username:
            return False
        number = self.milestone.number if self.milestone else None
        labels = [str(l) for l in self.original_labels]
        return self.edit(self.title, self.body, username, self.state, number,
                         labels)

    @requires_auth
    def close(self):
        """Close this issue.

        :returns: bool
        """
        assignee = self.assignee.login if self.assignee else ''
        number = self.milestone.number if self.milestone else None
        labels = [str(l) for l in self.original_labels]
        return self.edit(self.title, self.body, assignee, 'closed',
                         number, labels)

    def comment(self, id_num):
        """Get a single comment by its id.

        The catch here is that id is NOT a simple number to obtain. If
        you were to look at the comments on issue #15 in
        sigmavirus24/Todo.txt-python, the first comment's id is 4150787.

        :param int id_num: (required), comment id, see example above
        :returns: :class:`IssueComment <github3.issues.comment.IssueComment>`
        """
        json = None
        if int(id_num) > 0:  # Might as well check that it's positive
            owner, repo = self.repository
            url = self._build_url('repos', owner, repo, 'issues', 'comments',
                                  str(id_num))
            json = self._json(self._get(url), 200)
        return self._instance_or_null(IssueComment, json)

    def comments(self, number=-1, sort='', direction='', since=None):
        """Iterate over the comments on this issue.

        :param int number: (optional), number of comments to iterate over
            Default: -1 returns all comments
        :param str sort: accepted valuees: ('created', 'updated')
            api-default: created
        :param str direction: accepted values: ('asc', 'desc')
            Ignored without the sort parameter
        :param since: (optional), Only issues after this date will
            be returned. This can be a `datetime` or an ISO8601 formatted
            date string, e.g., 2012-05-20T23:10:27Z
        :type since: datetime or string
        :returns: iterator of
            :class:`IssueComment <github3.issues.comment.IssueComment>`\ s
        """
        url = self._build_url('comments', base_url=self._api)
        params = issue_comment_params(sort, direction, since)
        return self._iter(int(number), url, IssueComment, params)

    @requires_auth
    def create_comment(self, body):
        """Create a comment on this issue.

        :param str body: (required), comment body
        :returns: :class:`IssueComment <github3.issues.comment.IssueComment>`
        """
        json = None
        if body:
            url = self._build_url('comments', base_url=self._api)
            json = self._json(self._post(url, data={'body': body}),
                              201)
        return self._instance_or_null(IssueComment, json)

    @requires_auth
    def edit(self, title=None, body=None, assignee=None, state=None,
             milestone=None, labels=None):
        """Edit this issue.

        :param str title: Title of the issue
        :param str body: markdown formatted body (description) of the issue
        :param str assignee: login name of user the issue should be assigned
            to
        :param str state: accepted values: ('open', 'closed')
        :param int milestone: the NUMBER (not title) of the milestone to
            assign this to [1]_, or 0 to remove the milestone
        :param list labels: list of labels to apply this to
        :returns: bool

        .. [1] Milestone numbering starts at 1, i.e. the first milestone you
               create is 1, the second is 2, etc.
        """
        json = None
        data = {'title': title, 'body': body, 'assignee': assignee,
                'state': state, 'milestone': milestone, 'labels': labels}
        self._remove_none(data)
        if data:
            if 'milestone' in data and data['milestone'] == 0:
                data['milestone'] = None
            json = self._json(self._patch(self._api, data=dumps(data)), 200)
        if json:
            self._update_attributes(json)
            return True
        return False

    def events(self, number=-1):
        """Iterate over events associated with this issue only.

        :param int number: (optional), number of events to return. Default: -1
            returns all events available.
        :returns: generator of
            :class:`IssueEvent <github3.issues.event.IssueEvent>`\ s
        """
        url = self._build_url('events', base_url=self._api)
        return self._iter(int(number), url, IssueEvent)

    def is_closed(self):
        """Checks if the issue is closed.

        :returns: bool
        """
        if self.closed_at or (self.state == 'closed'):
            return True
        return False

    def labels(self, number=-1, etag=None):
        """Iterate over the labels associated with this issue.

        :param int number: (optional), number of labels to return. Default: -1
            returns all labels applied to this issue.
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Label <github3.issues.label.Label>`\ s
        """
        url = self._build_url('labels', base_url=self._api)
        return self._iter(int(number), url, Label, etag=etag)

    @requires_auth
    def lock(self):
        """Lock an issue.

        :returns: bool
        """
        headers = Issue.LOCKING_PREVIEW_HEADERS

        url = self._build_url('lock', base_url=self._api)
        return self._boolean(self._put(url, headers=headers), 204, 404)

    def pull_request(self):
        """Retrieve the pull request associated with this issue.

        :returns: :class:`~github3.pulls.PullRequest`
        """
        from .. import pulls
        json = None
        pull_request_url = self.pull_request_urls.get('url')
        if pull_request_url:
            json = self._json(self._get(pull_request_url), 200)
        return self._instance_or_null(pulls.PullRequest, json)

    @requires_auth
    def remove_label(self, name):
        """Removes label ``name`` from this issue.

        :param str name: (required), name of the label to remove
        :returns: list of :class:`Label`
        """
        url = self._build_url('labels', name, base_url=self._api)
        json = self._json(self._delete(url), 200, 404)
        labels = [Label(label, self) for label in json] if json else []
        return labels

    @requires_auth
    def remove_all_labels(self):
        """Remove all labels from this issue.

        :returns: an empty list if successful
        """
        # Can either send DELETE or [] to remove all labels
        return self.replace_labels([])

    @requires_auth
    def replace_labels(self, labels):
        """Replace all labels on this issue with ``labels``.

        :param list labels: label names
        :returns: list of :class:`Label`
        """
        url = self._build_url('labels', base_url=self._api)
        json = self._json(self._put(url, data=dumps(labels)), 200)
        return [Label(l, self) for l in json] if json else []

    @requires_auth
    def reopen(self):
        """Re-open a closed issue.

        :returns: bool
        """
        assignee = self.assignee.login if self.assignee else ''
        number = self.milestone.number if self.milestone else None
        labels = [str(l) for l in self.original_labels]
        return self.edit(self.title, self.body, assignee, 'open',
                         number, labels)

    @requires_auth
    def unlock(self):
        """Unlock an issue.

        :returns: bool
        """
        headers = Issue.LOCKING_PREVIEW_HEADERS

        url = self._build_url('lock', base_url=self._api)
        return self._boolean(self._delete(url, headers=headers), 204, 404)



"""Tools for helping with ANSI color codes."""
import re
import sys
import warnings
import builtins

from xonsh.platform import HAS_PYGMENTS
from xonsh.lazyasd import LazyDict, lazyobject
from xonsh.color_tools import (
    RE_BACKGROUND,
    BASE_XONSH_COLORS,
    make_palette,
    find_closest_color,
    rgb2short,
    rgb_to_256,
    short_to_ints,
)
from xonsh.tools import FORMATTER


def ansi_partial_color_format(template, style="default", cmap=None, hide=False):
    """Formats a template string but only with respect to the colors.
    Another template string is returned, with the color values filled in.

    Parameters
    ----------
    template : str
        The template string, potentially with color names.
    style : str, optional
        Style name to look up color map from.
    cmap : dict, optional
        A color map to use, this will prevent the color map from being
        looked up via the style name.
    hide : bool, optional
        Whether to wrap the color codes in the \\001 and \\002 escape
        codes, so that the color codes are not counted against line
        length.

    Returns
    -------
    A template string with the color values filled in.
    """
    try:
        return _ansi_partial_color_format_main(
            template, style=style, cmap=cmap, hide=hide
        )
    except Exception:
        return template


def _ansi_partial_color_format_main(template, style="default", cmap=None, hide=False):
    if cmap is not None:
        pass
    elif style in ANSI_STYLES:
        cmap = ANSI_STYLES[style]
    else:
        try:  # dynamically loading the style
            cmap = ansi_style_by_name(style)
        except Exception:
            msg = "Could not find color style {0!r}, using default."
            print(msg.format(style), file=sys.stderr)
            builtins.__xonsh__.env["XONSH_COLOR_STYLE"] = "default"
            cmap = ANSI_STYLES["default"]
    esc = ("\001" if hide else "") + "\033["
    m = "m" + ("\002" if hide else "")
    bopen = "{"
    bclose = "}"
    colon = ":"
    expl = "!"
    toks = []
    for literal, field, spec, conv in FORMATTER.parse(template):
        toks.append(literal)
        if field is None:
            pass
        elif field in cmap:
            toks.extend([esc, cmap[field], m])
        elif "#" in field:
            field = field.lower()
            pre, _, post = field.partition("#")
            f_or_b = "38" if RE_BACKGROUND.search(pre) is None else "48"
            rgb, _, post = post.partition("_")
            c256, _ = rgb_to_256(rgb)
            color = f_or_b + ";5;" + c256
            mods = pre + "_" + post
            if "underline" in mods:
                color = "4;" + color
            if "bold" in mods:
                color = "1;" + color
            toks.extend([esc, color, m])
        elif field is not None:
            toks.append(bopen)
            toks.append(field)
            if conv is not None and len(conv) > 0:
                toks.append(expl)
                toks.append(conv)
            if spec is not None and len(spec) > 0:
                toks.append(colon)
                toks.append(spec)
            toks.append(bclose)
    return "".join(toks)


def ansi_color_style_names():
    """Returns an iterable of all ANSI color style names."""
    return ANSI_STYLES.keys()


def ansi_color_style(style="default"):
    """Returns the current color map."""
    if style in ANSI_STYLES:
        cmap = ANSI_STYLES[style]
    else:
        msg = "Could not find color style {0!r}, using default.".format(style)
        warnings.warn(msg, RuntimeWarning)
        cmap = ANSI_STYLES["default"]
    return cmap


def ansi_reverse_style(style="default", return_style=False):
    """Reverses an ANSI color style mapping so that escape codes map to
    colors. Style may either be string or mapping. May also return
    the style it looked up.
    """
    style = ansi_style_by_name(style) if isinstance(style, str) else style
    reversed_style = {v: k for k, v in style.items()}
    # add keys to make this more useful
    updates = {
        "1": "BOLD_",
        "2": "FAINT_",
        "4": "UNDERLINE_",
        "5": "SLOWBLINK_",
        "1;4": "BOLD_UNDERLINE_",
        "4;1": "BOLD_UNDERLINE_",
        "38": "SET_FOREGROUND_",
        "48": "SET_BACKGROUND_",
        "38;2": "SET_FOREGROUND_3INTS_",
        "48;2": "SET_BACKGROUND_3INTS_",
        "38;5": "SET_FOREGROUND_SHORT_",
        "48;5": "SET_BACKGROUND_SHORT_",
    }
    for ec, name in reversed_style.items():
        no_left_zero = ec.lstrip("0")
        if no_left_zero.startswith(";"):
            updates[no_left_zero[1:]] = name
        elif no_left_zero != ec:
            updates[no_left_zero] = name
    reversed_style.update(updates)
    # return results
    if return_style:
        return style, reversed_style
    else:
        return reversed_style


@lazyobject
def ANSI_ESCAPE_CODE_RE():
    return re.compile(r"\001?(\033\[)?([0-9;]+)m?\002?")


@lazyobject
def ANSI_REVERSE_COLOR_NAME_TRANSLATIONS():
    base = {
        "SET_FOREGROUND_FAINT_": "SET_FOREGROUND_3INTS_",
        "SET_BACKGROUND_FAINT_": "SET_BACKGROUND_3INTS_",
        "SET_FOREGROUND_SLOWBLINK_": "SET_FOREGROUND_SHORT_",
        "SET_BACKGROUND_SLOWBLINK_": "SET_BACKGROUND_SHORT_",
    }
    data = {"UNDERLINE_BOLD_": "BOLD_UNDERLINE_"}
    data.update(base)
    data.update({"BOLD_" + k: "BOLD_" + v for k, v in base.items()})
    data.update({"UNDERLINE_" + k: "UNDERLINE_" + v for k, v in base.items()})
    data.update({"BOLD_UNDERLINE_" + k: "BOLD_UNDERLINE_" + v for k, v in base.items()})
    data.update({"UNDERLINE_BOLD_" + k: "BOLD_UNDERLINE_" + v for k, v in base.items()})
    return data


@lazyobject
def ANSI_COLOR_NAME_SET_3INTS_RE():
    return re.compile(r"(\w+_)?SET_(FORE|BACK)GROUND_3INTS_(\d+)_(\d+)_(\d+)")


@lazyobject
def ANSI_COLOR_NAME_SET_SHORT_RE():
    return re.compile(r"(\w+_)?SET_(FORE|BACK)GROUND_SHORT_(\d+)")


def _color_name_from_ints(ints, background=False, prefix=None):
    name = find_closest_color(ints, BASE_XONSH_COLORS)
    if background:
        name = "BACKGROUND_" + name
    name = name if prefix is None else prefix + name
    return name


_ANSI_COLOR_ESCAPE_CODE_TO_NAME_CACHE = {}


def ansi_color_escape_code_to_name(escape_code, style, reversed_style=None):
    """Converts an ASNI color code escape sequence to a tuple of color names
    in the provided style ('default' should almost be the style). For example,
    '0' becomes ('NO_COLOR',) and '32;41' becomes ('GREEN', 'BACKGROUND_RED').
    The style keyword may either be a string, in which the style is looked up,
    or an actual style dict.  You can also provide a reversed style mapping,
    too, which is just the keys/values of the style dict swapped. If reversed
    style is not provided, it is computed.
    """
    key = (escape_code, style)
    if key in _ANSI_COLOR_ESCAPE_CODE_TO_NAME_CACHE:
        return _ANSI_COLOR_ESCAPE_CODE_TO_NAME_CACHE[key]
    if reversed_style is None:
        style, reversed_style = ansi_reverse_style(style, return_style=True)
    # strip some actual escape codes, if needed.
    ec = ANSI_ESCAPE_CODE_RE.match(escape_code).group(2)
    names = []
    n_ints = 0
    seen_set_foreback = False
    for e in ec.split(";"):
        no_left_zero = e.lstrip("0") if len(e) > 1 else e
        if seen_set_foreback and n_ints > 0:
            names.append(e)
            n_ints -= 1
            if n_ints == 0:
                seen_set_foreback = False
            continue
        else:
            names.append(reversed_style.get(no_left_zero, no_left_zero))
        # set the flags for next time
        if "38" == e or "48" == e:
            seen_set_foreback = True
        elif "2" == e:
            n_ints = 3
        elif "5" == e:
            n_ints = 1
    # normalize names
    n = ""
    norm_names = []
    colors = set(reversed_style.values())
    for name in names:
        if name == "NO_COLOR":
            # skip most '0' entries
            continue
        n = n + name if n else name
        n = ANSI_REVERSE_COLOR_NAME_TRANSLATIONS.get(n, n)
        if n.endswith("_"):
            continue
        elif ANSI_COLOR_NAME_SET_SHORT_RE.match(n) is not None:
            pre, fore_back, short = ANSI_COLOR_NAME_SET_SHORT_RE.match(n).groups()
            n = _color_name_from_ints(
                short_to_ints(short), background=(fore_back == "BACK"), prefix=pre
            )
        elif ANSI_COLOR_NAME_SET_3INTS_RE.match(n) is not None:
            pre, fore_back, r, g, b = ANSI_COLOR_NAME_SET_3INTS_RE.match(n).groups()
            n = _color_name_from_ints(
                (int(r), int(g), int(b)), background=(fore_back == "BACK"), prefix=pre
            )
        elif "GROUND_3INTS_" in n:
            # have 1 or 2, but not 3 ints
            n += "_"
            continue
        # error check
        if n not in colors:
            msg = (
                "Could not translate ANSI color code {escape_code!r} "
                "into a known color in the palette. Specifically, the {n!r} "
                "portion of {name!r} in {names!r} seems to missing."
            )
            raise ValueError(
                msg.format(escape_code=escape_code, names=names, name=name, n=n)
            )
        norm_names.append(n)
        n = ""
    # return
    if len(norm_names) == 0:
        return ("NO_COLOR",)
    else:
        return tuple(norm_names)


def _ansi_expand_style(cmap):
    """Expands a style in order to more quickly make color map changes."""
    for key, val in list(cmap.items()):
        if key == "NO_COLOR":
            continue
        elif len(val) == 0:
            cmap["BOLD_" + key] = "1"
            cmap["UNDERLINE_" + key] = "4"
            cmap["BOLD_UNDERLINE_" + key] = "1;4"
            cmap["BACKGROUND_" + key] = val
        else:
            cmap["BOLD_" + key] = "1;" + val
            cmap["UNDERLINE_" + key] = "4;" + val
            cmap["BOLD_UNDERLINE_" + key] = "1;4;" + val
            cmap["BACKGROUND_" + key] = val.replace("38", "48", 1)


def _bw_style():
    style = {
        "BLACK": "0;30",
        "BLUE": "0;37",
        "CYAN": "0;37",
        "GREEN": "0;37",
        "INTENSE_BLACK": "0;90",
        "INTENSE_BLUE": "0;97",
        "INTENSE_CYAN": "0;97",
        "INTENSE_GREEN": "0;97",
        "INTENSE_PURPLE": "0;97",
        "INTENSE_RED": "0;97",
        "INTENSE_WHITE": "0;97",
        "INTENSE_YELLOW": "0;97",
        "NO_COLOR": "0",
        "PURPLE": "0;37",
        "RED": "0;37",
        "WHITE": "0;37",
        "YELLOW": "0;37",
    }
    _ansi_expand_style(style)
    return style


def _default_style():
    style = {
        # Reset
        "NO_COLOR": "0",  # Text Reset
        # Regular Colors
        "BLACK": "0;30",  # BLACK
        "RED": "0;31",  # RED
        "GREEN": "0;32",  # GREEN
        "YELLOW": "0;33",  # YELLOW
        "BLUE": "0;34",  # BLUE
        "PURPLE": "0;35",  # PURPLE
        "CYAN": "0;36",  # CYAN
        "WHITE": "0;37",  # WHITE
        # Bold
        "BOLD_BLACK": "1;30",  # BLACK
        "BOLD_RED": "1;31",  # RED
        "BOLD_GREEN": "1;32",  # GREEN
        "BOLD_YELLOW": "1;33",  # YELLOW
        "BOLD_BLUE": "1;34",  # BLUE
        "BOLD_PURPLE": "1;35",  # PURPLE
        "BOLD_CYAN": "1;36",  # CYAN
        "BOLD_WHITE": "1;37",  # WHITE
        # Underline
        "UNDERLINE_BLACK": "4;30",  # BLACK
        "UNDERLINE_RED": "4;31",  # RED
        "UNDERLINE_GREEN": "4;32",  # GREEN
        "UNDERLINE_YELLOW": "4;33",  # YELLOW
        "UNDERLINE_BLUE": "4;34",  # BLUE
        "UNDERLINE_PURPLE": "4;35",  # PURPLE
        "UNDERLINE_CYAN": "4;36",  # CYAN
        "UNDERLINE_WHITE": "4;37",  # WHITE
        # Bold, Underline
        "BOLD_UNDERLINE_BLACK": "1;4;30",  # BLACK
        "BOLD_UNDERLINE_RED": "1;4;31",  # RED
        "BOLD_UNDERLINE_GREEN": "1;4;32",  # GREEN
        "BOLD_UNDERLINE_YELLOW": "1;4;33",  # YELLOW
        "BOLD_UNDERLINE_BLUE": "1;4;34",  # BLUE
        "BOLD_UNDERLINE_PURPLE": "1;4;35",  # PURPLE
        "BOLD_UNDERLINE_CYAN": "1;4;36",  # CYAN
        "BOLD_UNDERLINE_WHITE": "1;4;37",  # WHITE
        # Background
        "BACKGROUND_BLACK": "40",  # BLACK
        "BACKGROUND_RED": "41",  # RED
        "BACKGROUND_GREEN": "42",  # GREEN
        "BACKGROUND_YELLOW": "43",  # YELLOW
        "BACKGROUND_BLUE": "44",  # BLUE
        "BACKGROUND_PURPLE": "45",  # PURPLE
        "BACKGROUND_CYAN": "46",  # CYAN
        "BACKGROUND_WHITE": "47",  # WHITE
        # High Intensity
        "INTENSE_BLACK": "0;90",  # BLACK
        "INTENSE_RED": "0;91",  # RED
        "INTENSE_GREEN": "0;92",  # GREEN
        "INTENSE_YELLOW": "0;93",  # YELLOW
        "INTENSE_BLUE": "0;94",  # BLUE
        "INTENSE_PURPLE": "0;95",  # PURPLE
        "INTENSE_CYAN": "0;96",  # CYAN
        "INTENSE_WHITE": "0;97",  # WHITE
        # Bold High Intensity
        "BOLD_INTENSE_BLACK": "1;90",  # BLACK
        "BOLD_INTENSE_RED": "1;91",  # RED
        "BOLD_INTENSE_GREEN": "1;92",  # GREEN
        "BOLD_INTENSE_YELLOW": "1;93",  # YELLOW
        "BOLD_INTENSE_BLUE": "1;94",  # BLUE
        "BOLD_INTENSE_PURPLE": "1;95",  # PURPLE
        "BOLD_INTENSE_CYAN": "1;96",  # CYAN
        "BOLD_INTENSE_WHITE": "1;97",  # WHITE
        # Underline High Intensity
        "UNDERLINE_INTENSE_BLACK": "4;90",  # BLACK
        "UNDERLINE_INTENSE_RED": "4;91",  # RED
        "UNDERLINE_INTENSE_GREEN": "4;92",  # GREEN
        "UNDERLINE_INTENSE_YELLOW": "4;93",  # YELLOW
        "UNDERLINE_INTENSE_BLUE": "4;94",  # BLUE
        "UNDERLINE_INTENSE_PURPLE": "4;95",  # PURPLE
        "UNDERLINE_INTENSE_CYAN": "4;96",  # CYAN
        "UNDERLINE_INTENSE_WHITE": "4;97",  # WHITE
        # Bold Underline High Intensity
        "BOLD_UNDERLINE_INTENSE_BLACK": "1;4;90",  # BLACK
        "BOLD_UNDERLINE_INTENSE_RED": "1;4;91",  # RED
        "BOLD_UNDERLINE_INTENSE_GREEN": "1;4;92",  # GREEN
        "BOLD_UNDERLINE_INTENSE_YELLOW": "1;4;93",  # YELLOW
        "BOLD_UNDERLINE_INTENSE_BLUE": "1;4;94",  # BLUE
        "BOLD_UNDERLINE_INTENSE_PURPLE": "1;4;95",  # PURPLE
        "BOLD_UNDERLINE_INTENSE_CYAN": "1;4;96",  # CYAN
        "BOLD_UNDERLINE_INTENSE_WHITE": "1;4;97",  # WHITE
        # High Intensity backgrounds
        "BACKGROUND_INTENSE_BLACK": "0;100",  # BLACK
        "BACKGROUND_INTENSE_RED": "0;101",  # RED
        "BACKGROUND_INTENSE_GREEN": "0;102",  # GREEN
        "BACKGROUND_INTENSE_YELLOW": "0;103",  # YELLOW
        "BACKGROUND_INTENSE_BLUE": "0;104",  # BLUE
        "BACKGROUND_INTENSE_PURPLE": "0;105",  # PURPLE
        "BACKGROUND_INTENSE_CYAN": "0;106",  # CYAN
        "BACKGROUND_INTENSE_WHITE": "0;107",  # WHITE
    }
    return style


def _monokai_style():
    style = {
        "NO_COLOR": "0",
        "BLACK": "38;5;16",
        "BLUE": "38;5;63",
        "CYAN": "38;5;81",
        "GREEN": "38;5;40",
        "PURPLE": "38;5;89",
        "RED": "38;5;124",
        "WHITE": "38;5;188",
        "YELLOW": "38;5;184",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;20",
        "INTENSE_CYAN": "38;5;44",
        "INTENSE_GREEN": "38;5;148",
        "INTENSE_PURPLE": "38;5;141",
        "INTENSE_RED": "38;5;197",
        "INTENSE_WHITE": "38;5;15",
        "INTENSE_YELLOW": "38;5;186",
    }
    _ansi_expand_style(style)
    return style


####################################
# Auto-generated below this line   #
####################################


def _algol_style():
    style = {
        "BLACK": "38;5;59",
        "BLUE": "38;5;59",
        "CYAN": "38;5;59",
        "GREEN": "38;5;59",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;102",
        "INTENSE_CYAN": "38;5;102",
        "INTENSE_GREEN": "38;5;102",
        "INTENSE_PURPLE": "38;5;102",
        "INTENSE_RED": "38;5;09",
        "INTENSE_WHITE": "38;5;102",
        "INTENSE_YELLOW": "38;5;102",
        "NO_COLOR": "0",
        "PURPLE": "38;5;59",
        "RED": "38;5;09",
        "WHITE": "38;5;102",
        "YELLOW": "38;5;09",
    }
    _ansi_expand_style(style)
    return style


def _algol_nu_style():
    style = {
        "BLACK": "38;5;59",
        "BLUE": "38;5;59",
        "CYAN": "38;5;59",
        "GREEN": "38;5;59",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;102",
        "INTENSE_CYAN": "38;5;102",
        "INTENSE_GREEN": "38;5;102",
        "INTENSE_PURPLE": "38;5;102",
        "INTENSE_RED": "38;5;09",
        "INTENSE_WHITE": "38;5;102",
        "INTENSE_YELLOW": "38;5;102",
        "NO_COLOR": "0",
        "PURPLE": "38;5;59",
        "RED": "38;5;09",
        "WHITE": "38;5;102",
        "YELLOW": "38;5;09",
    }
    _ansi_expand_style(style)
    return style


def _autumn_style():
    style = {
        "BLACK": "38;5;18",
        "BLUE": "38;5;19",
        "CYAN": "38;5;37",
        "GREEN": "38;5;34",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;33",
        "INTENSE_CYAN": "38;5;33",
        "INTENSE_GREEN": "38;5;64",
        "INTENSE_PURPLE": "38;5;217",
        "INTENSE_RED": "38;5;130",
        "INTENSE_WHITE": "38;5;145",
        "INTENSE_YELLOW": "38;5;217",
        "NO_COLOR": "0",
        "PURPLE": "38;5;90",
        "RED": "38;5;124",
        "WHITE": "38;5;145",
        "YELLOW": "38;5;130",
    }
    _ansi_expand_style(style)
    return style


def _borland_style():
    style = {
        "BLACK": "38;5;16",
        "BLUE": "38;5;18",
        "CYAN": "38;5;30",
        "GREEN": "38;5;28",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;21",
        "INTENSE_CYAN": "38;5;194",
        "INTENSE_GREEN": "38;5;102",
        "INTENSE_PURPLE": "38;5;188",
        "INTENSE_RED": "38;5;09",
        "INTENSE_WHITE": "38;5;224",
        "INTENSE_YELLOW": "38;5;188",
        "NO_COLOR": "0",
        "PURPLE": "38;5;90",
        "RED": "38;5;124",
        "WHITE": "38;5;145",
        "YELLOW": "38;5;124",
    }
    _ansi_expand_style(style)
    return style


def _colorful_style():
    style = {
        "BLACK": "38;5;16",
        "BLUE": "38;5;20",
        "CYAN": "38;5;31",
        "GREEN": "38;5;34",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;61",
        "INTENSE_CYAN": "38;5;145",
        "INTENSE_GREEN": "38;5;102",
        "INTENSE_PURPLE": "38;5;217",
        "INTENSE_RED": "38;5;166",
        "INTENSE_WHITE": "38;5;15",
        "INTENSE_YELLOW": "38;5;217",
        "NO_COLOR": "0",
        "PURPLE": "38;5;90",
        "RED": "38;5;124",
        "WHITE": "38;5;145",
        "YELLOW": "38;5;130",
    }
    _ansi_expand_style(style)
    return style


def _emacs_style():
    style = {
        "BLACK": "38;5;28",
        "BLUE": "38;5;18",
        "CYAN": "38;5;26",
        "GREEN": "38;5;34",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;26",
        "INTENSE_CYAN": "38;5;145",
        "INTENSE_GREEN": "38;5;34",
        "INTENSE_PURPLE": "38;5;129",
        "INTENSE_RED": "38;5;167",
        "INTENSE_WHITE": "38;5;145",
        "INTENSE_YELLOW": "38;5;145",
        "NO_COLOR": "0",
        "PURPLE": "38;5;90",
        "RED": "38;5;124",
        "WHITE": "38;5;145",
        "YELLOW": "38;5;130",
    }
    _ansi_expand_style(style)
    return style


def _friendly_style():
    style = {
        "BLACK": "38;5;22",
        "BLUE": "38;5;18",
        "CYAN": "38;5;31",
        "GREEN": "38;5;34",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;74",
        "INTENSE_CYAN": "38;5;74",
        "INTENSE_GREEN": "38;5;71",
        "INTENSE_PURPLE": "38;5;134",
        "INTENSE_RED": "38;5;167",
        "INTENSE_WHITE": "38;5;15",
        "INTENSE_YELLOW": "38;5;145",
        "NO_COLOR": "0",
        "PURPLE": "38;5;90",
        "RED": "38;5;124",
        "WHITE": "38;5;145",
        "YELLOW": "38;5;166",
    }
    _ansi_expand_style(style)
    return style


def _fruity_style():
    style = {
        "BLACK": "38;5;16",
        "BLUE": "38;5;32",
        "CYAN": "38;5;32",
        "GREEN": "38;5;28",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;33",
        "INTENSE_CYAN": "38;5;33",
        "INTENSE_GREEN": "38;5;102",
        "INTENSE_PURPLE": "38;5;198",
        "INTENSE_RED": "38;5;202",
        "INTENSE_WHITE": "38;5;15",
        "INTENSE_YELLOW": "38;5;187",
        "NO_COLOR": "0",
        "PURPLE": "38;5;198",
        "RED": "38;5;09",
        "WHITE": "38;5;187",
        "YELLOW": "38;5;202",
    }
    _ansi_expand_style(style)
    return style


def _igor_style():
    style = {
        "BLACK": "38;5;34",
        "BLUE": "38;5;21",
        "CYAN": "38;5;30",
        "GREEN": "38;5;34",
        "INTENSE_BLACK": "38;5;30",
        "INTENSE_BLUE": "38;5;21",
        "INTENSE_CYAN": "38;5;30",
        "INTENSE_GREEN": "38;5;34",
        "INTENSE_PURPLE": "38;5;163",
        "INTENSE_RED": "38;5;166",
        "INTENSE_WHITE": "38;5;163",
        "INTENSE_YELLOW": "38;5;166",
        "NO_COLOR": "0",
        "PURPLE": "38;5;163",
        "RED": "38;5;166",
        "WHITE": "38;5;163",
        "YELLOW": "38;5;166",
    }
    _ansi_expand_style(style)
    return style


def _lovelace_style():
    style = {
        "BLACK": "38;5;59",
        "BLUE": "38;5;25",
        "CYAN": "38;5;29",
        "GREEN": "38;5;65",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;25",
        "INTENSE_CYAN": "38;5;102",
        "INTENSE_GREEN": "38;5;29",
        "INTENSE_PURPLE": "38;5;133",
        "INTENSE_RED": "38;5;131",
        "INTENSE_WHITE": "38;5;102",
        "INTENSE_YELLOW": "38;5;136",
        "NO_COLOR": "0",
        "PURPLE": "38;5;133",
        "RED": "38;5;124",
        "WHITE": "38;5;102",
        "YELLOW": "38;5;130",
    }
    _ansi_expand_style(style)
    return style


def _manni_style():
    style = {
        "BLACK": "38;5;16",
        "BLUE": "38;5;18",
        "CYAN": "38;5;30",
        "GREEN": "38;5;40",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;105",
        "INTENSE_CYAN": "38;5;45",
        "INTENSE_GREEN": "38;5;113",
        "INTENSE_PURPLE": "38;5;165",
        "INTENSE_RED": "38;5;202",
        "INTENSE_WHITE": "38;5;224",
        "INTENSE_YELLOW": "38;5;221",
        "NO_COLOR": "0",
        "PURPLE": "38;5;165",
        "RED": "38;5;124",
        "WHITE": "38;5;145",
        "YELLOW": "38;5;166",
    }
    _ansi_expand_style(style)
    return style


def _murphy_style():
    style = {
        "BLACK": "38;5;16",
        "BLUE": "38;5;18",
        "CYAN": "38;5;31",
        "GREEN": "38;5;34",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;63",
        "INTENSE_CYAN": "38;5;86",
        "INTENSE_GREEN": "38;5;86",
        "INTENSE_PURPLE": "38;5;213",
        "INTENSE_RED": "38;5;209",
        "INTENSE_WHITE": "38;5;15",
        "INTENSE_YELLOW": "38;5;222",
        "NO_COLOR": "0",
        "PURPLE": "38;5;90",
        "RED": "38;5;124",
        "WHITE": "38;5;145",
        "YELLOW": "38;5;166",
    }
    _ansi_expand_style(style)
    return style


def _native_style():
    style = {
        "BLACK": "38;5;52",
        "BLUE": "38;5;67",
        "CYAN": "38;5;31",
        "GREEN": "38;5;64",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;68",
        "INTENSE_CYAN": "38;5;87",
        "INTENSE_GREEN": "38;5;70",
        "INTENSE_PURPLE": "38;5;188",
        "INTENSE_RED": "38;5;160",
        "INTENSE_WHITE": "38;5;15",
        "INTENSE_YELLOW": "38;5;214",
        "NO_COLOR": "0",
        "PURPLE": "38;5;59",
        "RED": "38;5;124",
        "WHITE": "38;5;145",
        "YELLOW": "38;5;124",
    }
    _ansi_expand_style(style)
    return style


def _paraiso_dark_style():
    style = {
        "BLACK": "38;5;95",
        "BLUE": "38;5;97",
        "CYAN": "38;5;39",
        "GREEN": "38;5;72",
        "INTENSE_BLACK": "38;5;95",
        "INTENSE_BLUE": "38;5;97",
        "INTENSE_CYAN": "38;5;79",
        "INTENSE_GREEN": "38;5;72",
        "INTENSE_PURPLE": "38;5;188",
        "INTENSE_RED": "38;5;203",
        "INTENSE_WHITE": "38;5;188",
        "INTENSE_YELLOW": "38;5;220",
        "NO_COLOR": "0",
        "PURPLE": "38;5;97",
        "RED": "38;5;203",
        "WHITE": "38;5;79",
        "YELLOW": "38;5;214",
    }
    _ansi_expand_style(style)
    return style


def _paraiso_light_style():
    style = {
        "BLACK": "38;5;16",
        "BLUE": "38;5;16",
        "CYAN": "38;5;39",
        "GREEN": "38;5;72",
        "INTENSE_BLACK": "38;5;16",
        "INTENSE_BLUE": "38;5;97",
        "INTENSE_CYAN": "38;5;79",
        "INTENSE_GREEN": "38;5;72",
        "INTENSE_PURPLE": "38;5;97",
        "INTENSE_RED": "38;5;203",
        "INTENSE_WHITE": "38;5;79",
        "INTENSE_YELLOW": "38;5;220",
        "NO_COLOR": "0",
        "PURPLE": "38;5;97",
        "RED": "38;5;16",
        "WHITE": "38;5;102",
        "YELLOW": "38;5;214",
    }
    _ansi_expand_style(style)
    return style


def _pastie_style():
    style = {
        "BLACK": "38;5;16",
        "BLUE": "38;5;20",
        "CYAN": "38;5;25",
        "GREEN": "38;5;28",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;61",
        "INTENSE_CYAN": "38;5;194",
        "INTENSE_GREEN": "38;5;34",
        "INTENSE_PURPLE": "38;5;188",
        "INTENSE_RED": "38;5;172",
        "INTENSE_WHITE": "38;5;15",
        "INTENSE_YELLOW": "38;5;188",
        "NO_COLOR": "0",
        "PURPLE": "38;5;125",
        "RED": "38;5;124",
        "WHITE": "38;5;145",
        "YELLOW": "38;5;130",
    }
    _ansi_expand_style(style)
    return style


def _perldoc_style():
    style = {
        "BLACK": "38;5;18",
        "BLUE": "38;5;18",
        "CYAN": "38;5;31",
        "GREEN": "38;5;34",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;134",
        "INTENSE_CYAN": "38;5;145",
        "INTENSE_GREEN": "38;5;28",
        "INTENSE_PURPLE": "38;5;134",
        "INTENSE_RED": "38;5;167",
        "INTENSE_WHITE": "38;5;188",
        "INTENSE_YELLOW": "38;5;188",
        "NO_COLOR": "0",
        "PURPLE": "38;5;90",
        "RED": "38;5;124",
        "WHITE": "38;5;145",
        "YELLOW": "38;5;166",
    }
    _ansi_expand_style(style)
    return style


def _rrt_style():
    style = {
        "BLACK": "38;5;09",
        "BLUE": "38;5;117",
        "CYAN": "38;5;117",
        "GREEN": "38;5;46",
        "INTENSE_BLACK": "38;5;117",
        "INTENSE_BLUE": "38;5;117",
        "INTENSE_CYAN": "38;5;122",
        "INTENSE_GREEN": "38;5;46",
        "INTENSE_PURPLE": "38;5;213",
        "INTENSE_RED": "38;5;09",
        "INTENSE_WHITE": "38;5;188",
        "INTENSE_YELLOW": "38;5;222",
        "NO_COLOR": "0",
        "PURPLE": "38;5;213",
        "RED": "38;5;09",
        "WHITE": "38;5;117",
        "YELLOW": "38;5;09",
    }
    _ansi_expand_style(style)
    return style


def _tango_style():
    style = {
        "BLACK": "38;5;16",
        "BLUE": "38;5;20",
        "CYAN": "38;5;61",
        "GREEN": "38;5;34",
        "INTENSE_BLACK": "38;5;24",
        "INTENSE_BLUE": "38;5;62",
        "INTENSE_CYAN": "38;5;15",
        "INTENSE_GREEN": "38;5;64",
        "INTENSE_PURPLE": "38;5;15",
        "INTENSE_RED": "38;5;09",
        "INTENSE_WHITE": "38;5;15",
        "INTENSE_YELLOW": "38;5;178",
        "NO_COLOR": "0",
        "PURPLE": "38;5;90",
        "RED": "38;5;124",
        "WHITE": "38;5;15",
        "YELLOW": "38;5;94",
    }
    _ansi_expand_style(style)
    return style


def _trac_style():
    style = {
        "BLACK": "38;5;16",
        "BLUE": "38;5;18",
        "CYAN": "38;5;30",
        "GREEN": "38;5;100",
        "INTENSE_BLACK": "38;5;59",
        "INTENSE_BLUE": "38;5;60",
        "INTENSE_CYAN": "38;5;194",
        "INTENSE_GREEN": "38;5;102",
        "INTENSE_PURPLE": "38;5;188",
        "INTENSE_RED": "38;5;137",
        "INTENSE_WHITE": "38;5;224",
        "INTENSE_YELLOW": "38;5;188",
        "NO_COLOR": "0",
        "PURPLE": "38;5;90",
        "RED": "38;5;124",
        "WHITE": "38;5;145",
        "YELLOW": "38;5;100",
    }
    _ansi_expand_style(style)
    return style


def _vim_style():
    style = {
        "BLACK": "38;5;18",
        "BLUE": "38;5;18",
        "CYAN": "38;5;44",
        "GREEN": "38;5;40",
        "INTENSE_BLACK": "38;5;60",
        "INTENSE_BLUE": "38;5;68",
        "INTENSE_CYAN": "38;5;44",
        "INTENSE_GREEN": "38;5;40",
        "INTENSE_PURPLE": "38;5;164",
        "INTENSE_RED": "38;5;09",
        "INTENSE_WHITE": "38;5;188",
        "INTENSE_YELLOW": "38;5;184",
        "NO_COLOR": "0",
        "PURPLE": "38;5;164",
        "RED": "38;5;160",
        "WHITE": "38;5;188",
        "YELLOW": "38;5;160",
    }
    _ansi_expand_style(style)
    return style


def _vs_style():
    style = {
        "BLACK": "38;5;28",
        "BLUE": "38;5;21",
        "CYAN": "38;5;31",
        "GREEN": "38;5;28",
        "INTENSE_BLACK": "38;5;31",
        "INTENSE_BLUE": "38;5;31",
        "INTENSE_CYAN": "38;5;31",
        "INTENSE_GREEN": "38;5;31",
        "INTENSE_PURPLE": "38;5;31",
        "INTENSE_RED": "38;5;09",
        "INTENSE_WHITE": "38;5;31",
        "INTENSE_YELLOW": "38;5;31",
        "NO_COLOR": "0",
        "PURPLE": "38;5;124",
        "RED": "38;5;124",
        "WHITE": "38;5;31",
        "YELLOW": "38;5;124",
    }
    _ansi_expand_style(style)
    return style


def _xcode_style():
    style = {
        "BLACK": "38;5;16",
        "BLUE": "38;5;20",
        "CYAN": "38;5;60",
        "GREEN": "38;5;28",
        "INTENSE_BLACK": "38;5;60",
        "INTENSE_BLUE": "38;5;20",
        "INTENSE_CYAN": "38;5;60",
        "INTENSE_GREEN": "38;5;60",
        "INTENSE_PURPLE": "38;5;126",
        "INTENSE_RED": "38;5;160",
        "INTENSE_WHITE": "38;5;60",
        "INTENSE_YELLOW": "38;5;94",
        "NO_COLOR": "0",
        "PURPLE": "38;5;126",
        "RED": "38;5;160",
        "WHITE": "38;5;60",
        "YELLOW": "38;5;94",
    }
    _ansi_expand_style(style)
    return style


ANSI_STYLES = LazyDict(
    {
        "algol": _algol_style,
        "algol_nu": _algol_nu_style,
        "autumn": _autumn_style,
        "borland": _borland_style,
        "bw": _bw_style,
        "colorful": _colorful_style,
        "default": _default_style,
        "emacs": _emacs_style,
        "friendly": _friendly_style,
        "fruity": _fruity_style,
        "igor": _igor_style,
        "lovelace": _lovelace_style,
        "manni": _manni_style,
        "monokai": _monokai_style,
        "murphy": _murphy_style,
        "native": _native_style,
        "paraiso-dark": _paraiso_dark_style,
        "paraiso-light": _paraiso_light_style,
        "pastie": _pastie_style,
        "perldoc": _perldoc_style,
        "rrt": _rrt_style,
        "tango": _tango_style,
        "trac": _trac_style,
        "vim": _vim_style,
        "vs": _vs_style,
        "xcode": _xcode_style,
    },
    globals(),
    "ANSI_STYLES",
)

del (
    _algol_style,
    _algol_nu_style,
    _autumn_style,
    _borland_style,
    _bw_style,
    _colorful_style,
    _default_style,
    _emacs_style,
    _friendly_style,
    _fruity_style,
    _igor_style,
    _lovelace_style,
    _manni_style,
    _monokai_style,
    _murphy_style,
    _native_style,
    _paraiso_dark_style,
    _paraiso_light_style,
    _pastie_style,
    _perldoc_style,
    _rrt_style,
    _tango_style,
    _trac_style,
    _vim_style,
    _vs_style,
    _xcode_style,
)


#
# Dynamically generated styles
#
def make_ansi_style(palette):
    """Makes an ANSI color style from a color palette"""
    style = {"NO_COLOR": "0"}
    for name, t in BASE_XONSH_COLORS.items():
        closest = find_closest_color(t, palette)
        if len(closest) == 3:
            closest = "".join([a * 2 for a in closest])
        short = rgb2short(closest)[0]
        style[name] = "38;5;" + short
        style["BOLD_" + name] = "1;38;5;" + short
        style["UNDERLINE_" + name] = "4;38;5;" + short
        style["BOLD_UNDERLINE_" + name] = "1;4;38;5;" + short
        style["BACKGROUND_" + name] = "48;5;" + short
    return style


def ansi_style_by_name(name):
    """Gets or makes an ANSI color style by name. If the styles does not
    exist, it will look for a style using the pygments name.
    """
    if name in ANSI_STYLES:
        return ANSI_STYLES[name]
    elif not HAS_PYGMENTS:
        raise KeyError("could not find style {0!r}".format(name))
    from xonsh.pygments_cache import get_style_by_name

    pstyle = get_style_by_name(name)
    palette = make_palette(pstyle.styles.values())
    astyle = make_ansi_style(palette)
    ANSI_STYLES[name] = astyle
    return astyle



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ..models import GitHubCore
from ..users import User


class Deployment(GitHubCore):
    def _update_attributes(self, deployment):
        self._api = deployment.get('url')

        #: GitHub's id of this deployment
        self.id = deployment.get('id')

        #: SHA of the branch on GitHub
        self.sha = deployment.get('sha')

        #: The reference used to create the Deployment, e.g.,
        #: `deploy-20140526`
        self.ref = deployment.get('ref')

        #: User object representing the creator of the deployment
        self.creator = deployment.get('creator')
        if self.creator:
            self.creator = User(self.creator, self)

        #: JSON string payload of the Deployment
        self.payload = deployment.get('payload')

        #: Date the Deployment was created
        self.created_at = self._strptime(deployment.get('created_at'))

        #: Date the Deployment was updated
        self.updated_at = self._strptime(deployment.get('updated_at'))

        #: Description of the deployment
        self.description = deployment.get('description')

        #: Target for the deployment, e.g., 'production', 'staging'
        self.environment = deployment.get('environment')

        #: URL to get the statuses of this deployment
        self.statuses_url = deployment.get('statuses_url')

    def _repr(self):
        return '<Deployment [{0} @ {1}]>'.format(self.id, self.sha)

    def create_status(self, state, target_url=None, description=None):
        """Create a new deployment status for this deployment.

        :param str state: (required), The state of the status. Can be one of
            ``pending``, ``success``, ``error``, or ``failure``.
        :param str target_url: The target URL to associate with this status.
            This URL should contain output to keep the user updated while the
            task is running or serve as historical information for what
            happened in the deployment. Default: ''.
        :param str description: A short description of the status. Default: ''.
        :return: partial :class:`DeploymentStatus <DeploymentStatus>`
        """
        json = None

        if state in ('pending', 'success', 'error', 'failure'):
            data = {'state': state, 'target_url': target_url,
                    'description': description}
            self._remove_none(data)
            response = self._post(self.statuses_url, data=data)
            json = self._json(response, 201)

        return self._instance_or_null(DeploymentStatus, json)

    def statuses(self, number=-1, etag=None):
        """Iterate over the deployment statuses for this deployment.

        :param int number: (optional), the number of statuses to return.
            Default: -1, returns all statuses.
        :param str etag: (optional), the ETag header value from the last time
            you iterated over the statuses.
        :returns: generator of :class:`DeploymentStatus`\ es
        """
        i = self._iter(int(number), self.statuses_url, DeploymentStatus,
                       etag=etag)
        return i


class DeploymentStatus(GitHubCore):
    def _update_attributes(self, status):
        self._api = status.get('url')

        #: GitHub's id for this deployment status
        self.id = status.get('id')

        #: State of the deployment status
        self.state = status.get('state')

        #: Creater of the deployment status
        self.creator = status.get('creator')
        if self.creator:
            self.creator = User(self.creator, self)

        #: JSON payload as a string
        self.payload = status.get('payload', {})

        #: Target URL of the deployment
        self.target_url = status.get('target_url')

        #: Date the deployment status was created
        self.created_at = self._strptime(status.get('created_at'))

        #: Date the deployment status was updated
        self.updated_at = self._strptime(status.get('updated_at'))

        #: Description of the deployment
        self.description = status.get('description')

        #: :class:`Deployment` representing the deployment this status is
        #: associated with
        self.deployment = status.get('deployment')
        if self.deployment:
            self.deployment = Deployment(self.deployment, self)

        #: URL for the deployment this status is associated with
        self.deployment_url = status.get('deployment_url')

    def _repr(self):
        return '<DeploymentStatus [{0}]>'.format(self.id)



# -*- coding: utf-8 -*-
"""A collection of useful utilities."""
import collections
import datetime
import re

from requests import compat

# with thanks to https://code.google.com/p/jquery-localtime/issues/detail?id=4
ISO_8601 = re.compile("^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[0-1]|0"
                      "[1-9]|[1-2][0-9])(T(2[0-3]|[0-1][0-9]):([0-5][0-9]):([0"
                      "-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[0-1][0-9]):[0-5]["
                      "0-9])?)?$")


def timestamp_parameter(timestamp, allow_none=True):
    """Function to check the conformance of timestamps passed by users.

    This will check that a string is a valid format and allow users to pass a
    datetime object which we will then convert to a proper ISO8601 date-time
    string.

    :param timestamp: string to be validated or datetime object to be
        converted.
    :param bool allow_none: whether or not to allow timestamp to be None.
        Default: ``True``
    :returns: valid ISO8601 string
    :rtype: str
    :raises: ValueError
    """
    if timestamp is None:
        if allow_none:
            return None
        raise ValueError("Timestamp value cannot be None")

    if isinstance(timestamp, datetime.datetime):
        return timestamp.isoformat() + 'Z'

    if isinstance(timestamp, compat.basestring):
        if not ISO_8601.match(timestamp):
            raise ValueError(("Invalid timestamp: %s is not a valid ISO-8601"
                              " formatted date") % timestamp)
        return timestamp

    raise ValueError("Cannot accept type %s for timestamp" % type(timestamp))


class UTC(datetime.tzinfo):

    """Yet another UTC reimplementation, to avoid a dependency on pytz or
    dateutil."""

    ZERO = datetime.timedelta(0)

    def __repr__(self):
        return 'UTC()'

    def dst(self, dt):
        return self.ZERO

    def tzname(self, dt):
        return 'UTC'

    def utcoffset(self, dt):
        return self.ZERO


def stream_response_to_file(response, path=None):
    """Stream a response body to the specified file.

    Either use the ``path`` provided or use the name provided in the
    ``Content-Disposition`` header.

    :param response: A Response object from requests
    :type response: requests.models.Response
    :param str path: The full path and file name used to save the response
    :return: path to the file
    :rtype: str
    """
    pre_opened = False
    fd = None
    filename = None
    if path:
        if isinstance(getattr(path, 'write', None), collections.Callable):
            pre_opened = True
            fd = path
            filename = getattr(fd, 'name', None)
        else:
            fd = open(path, 'wb')
            filename = path
    else:
        header = response.headers['content-disposition']
        i = header.find('filename=') + len('filename=')
        filename = header[i:]
        fd = open(filename, 'wb')

    for chunk in response.iter_content(chunk_size=512):
        fd.write(chunk)

    if not pre_opened:
        fd.close()

    return filename



# -*- coding: utf-8 -*-
"""All exceptions for the github3 library."""


class GitHubError(Exception):
    """The base exception class."""

    def __init__(self, resp):
        super(GitHubError, self).__init__(resp)
        #: Response code that triggered the error
        self.response = resp
        self.code = resp.status_code
        self.errors = []
        try:
            error = resp.json()
            #: Message associated with the error
            self.msg = error.get('message')
            #: List of errors provided by GitHub
            if error.get('errors'):
                self.errors = error.get('errors')
        except:  # Amazon S3 error
            self.msg = resp.content or '[No message]'

    def __repr__(self):
        return '<{0} [{1}]>'.format(self.__class__.__name__,
                                    self.msg or self.code)

    def __str__(self):
        return '{0} {1}'.format(self.code, self.msg)

    @property
    def message(self):
        """The actual message returned by the API."""
        return self.msg


class UnprocessableResponseBody(GitHubError):
    """Exception class for response objects that cannot be handled."""
    def __init__(self, message, body):
        Exception.__init__(self, message)
        self.body = body
        self.msg = message

    def __repr__(self):
        return '<{0} [{1}]>'.format('UnprocessableResponseBody', self.body)

    def __str__(self):
        return self.message


class BadRequest(GitHubError):
    """Exception class for 400 responses."""
    pass


class AuthenticationFailed(GitHubError):
    """Exception class for 401 responses.

    Possible reasons:

    - Need one time password (for two-factor authentication)
    - You are not authorized to access the resource
    """
    pass


class ForbiddenError(GitHubError):
    """Exception class for 403 responses.

    Possible reasons:

    - Too many requests (you've exceeded the ratelimit)
    - Too many login failures
    """
    pass


class NotFoundError(GitHubError):
    """Exception class for 404 responses."""
    pass


class MethodNotAllowed(GitHubError):
    """Exception class for 405 responses."""
    pass


class NotAcceptable(GitHubError):
    """Exception class for 406 responses."""
    pass


class UnprocessableEntity(GitHubError):
    """Exception class for 422 responses."""
    pass


class ClientError(GitHubError):
    """Catch-all for 400 responses that aren't specific errors."""
    pass


class ServerError(GitHubError):
    """Exception class for 5xx responses."""
    pass


error_classes = {
    400: BadRequest,
    401: AuthenticationFailed,
    403: ForbiddenError,
    404: NotFoundError,
    405: MethodNotAllowed,
    406: NotAcceptable,
    422: UnprocessableEntity,
}


def error_for(response):
    """Return the appropriate initialized exception class for a response."""
    klass = error_classes.get(response.status_code)
    if klass is None:
        if 400 <= response.status_code < 500:
            klass = ClientError
        if 500 <= response.status_code < 600:
            klass = ServerError
    return klass(response)



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

formatted_trends = u'\x1b[35m  1.   \x1b[0mpuikinsh/gentelella \x1b[0m\n         Free Bootstrap 3 Admin Template\x1b[0m\n         HTML\x1b[0m\n\x1b[35m  2.   \x1b[0mrg3/youtube-dl \x1b[0m\n         Small command-line program to download videos from YouTube.com and\n         other video sites\x1b[0m\n         Python\x1b[0m\n\x1b[35m  3.   \x1b[0mFreeCodeCamp/FreeCodeCamp \x1b[0m\n         The http://FreeCodeCamp.com open source codebase and curriculum.\n         Learn to code and help nonprofits.\x1b[0m\n         JavaScript\x1b[0m\n\x1b[35m  4.   \x1b[0mmartanne/vis \x1b[0m\n         a vim like text editor\x1b[0m\n         C\x1b[0m\n\x1b[35m  5.   \x1b[0mFezVrasta/popper.js \x1b[0m\n         A ~3.5KB kickass library to manage your poppers\x1b[0m\n         JavaScript\x1b[0m\n\x1b[35m  6.   \x1b[0mfirehol/netdata \x1b[0m\n         Real-time performance monitoring, done right!\x1b[0m\n         C\x1b[0m\n\x1b[35m  7.   \x1b[0marachnys/athenapdf \x1b[0m\n         Drop-in replacement for wkhtmltopdf built on Go, Electron and Docker\x1b[0m\n         Go\x1b[0m\n\x1b[35m  8.   \x1b[0mmortenjust/cleartext-mac \x1b[0m\n         A text editor that only allows the top 1,000 most common words in\n         English\x1b[0m\n         Swift\x1b[0m\n\x1b[35m  9.   \x1b[0mfeross/webtorrent-desktop \x1b[0m\n         WebTorrent, the streaming torrent client. For OS X, Windows, and\n         Linux.\x1b[0m\n         JavaScript\x1b[0m\n\x1b[35m  10.  \x1b[0mfuture-architect/vuls \x1b[0m\n         Vulnerability scanner for Linux, agentless, written in golang.\x1b[0m\n         Go\x1b[0m\n\x1b[35m  11.  \x1b[0mDevLight-Mobile-Agency/NavigationTabBar \x1b[0m\n         Navigation tab bar with colorful interactions.\x1b[0m\n         Java\x1b[0m\n\x1b[35m  12.  \x1b[0mfeross/webtorrent \x1b[0m\n         Streaming torrent client for the web\x1b[0m\n         JavaScript\x1b[0m\n\x1b[35m  13.  \x1b[0mfacebook/react \x1b[0m\n         A declarative, efficient, and flexible JavaScript library for\n         building user interfaces.\x1b[0m\n         JavaScript\x1b[0m\n\x1b[35m  14.  \x1b[0mtatsuhiro-t/aria2 \x1b[0m\n         aria2 is a lightweight multi-protocol & multi-source, cross platform\n         download utility operated in command-line. It supports HTTP/HTTPS,\n         FTP, SFTP, BitTorrent and Metalink.\x1b[0m\n         C++\x1b[0m\n\x1b[35m  15.  \x1b[0mjessepollak/command \x1b[0m\n         Making the web better with Slack-like slash commands.\x1b[0m\n         JavaScript\x1b[0m\n\x1b[35m  16.  \x1b[0msoimort/you-get \x1b[0m\n         Dumb downloader that scrapes the web\x1b[0m\n         Python\x1b[0m\n\x1b[35m  17.  \x1b[0mpacketzoom/logzoom \x1b[0m\n         A lightweight replacement for logstash indexer in Go\x1b[0m\n         Go\x1b[0m\n\x1b[35m  18.  \x1b[0mairbnb/caravel \x1b[0m\n         Caravel is a data exploration platform designed to be visual,\n         intuitive, and interactive\x1b[0m\n         Python\x1b[0m\n\x1b[35m  19.  \x1b[0mdylang/npm-check \x1b[0m\n         Check for outdated, incorrect, and unused dependencies.\x1b[0m\n         JavaScript\x1b[0m\n\x1b[35m  20.  \x1b[0mkadirahq/react-storybook \x1b[0m\n         Isolate your React UI Component development from the main app\x1b[0m\n         JavaScript\x1b[0m\n\x1b[35m  21.  \x1b[0mnnupoor/js_designpatterns \x1b[0m\x1b[0m\n         JavaScript\x1b[0m\n\x1b[35m  22.  \x1b[0mgooglesamples/android-architecture \x1b[0m\n         A collection of samples to discuss and showcase different\n         architectural tools and patterns for Android apps.\x1b[0m\n         Unknown\x1b[0m\n\x1b[35m  23.  \x1b[0mMrS0m30n3/youtube-dl-gui \x1b[0m\n         A cross platform front-end GUI of the popular youtube-dl written in\n         wxPython.\x1b[0m\n         Python\x1b[0m\n\x1b[35m  24.  \x1b[0mSkyscanner/SkyFloatingLabelTextField \x1b[0m\n         A beautiful and flexible text field control implementation of "Float\n         Label Pattern". Written in Swift. As used in the Skyscanner TravelPro\n         iOS app.\x1b[0m\n         Swift\x1b[0m\n\x1b[35m  25.  \x1b[0mshutterstock/List-of-Dirty-Naughty-Obscene-and-Otherwise-Bad-Words \x1b[0m\n         List of Dirty, Naughty, Obscene, and Otherwise Bad Words\x1b[0m\n         Unknown\x1b[0m\n  View the page for \x1b[0m\x1b[35m1 through \x1b[0m\x1b[35m25\x1b[0m with the following command:\n\x1b[0m\x1b[35m    gh view [#] \x1b[0moptional: [-b/--browser] [--help]\n\x1b[0m\x1b[0m'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from json import dumps
from ..decorators import requires_auth
from .label import Label
from ..models import GitHubCore
from ..users import User


class Milestone(GitHubCore):
    """The :class:`Milestone <Milestone>` object. This is a small class to
    handle information about milestones on repositories and issues.

    See also: http://developer.github.com/v3/issues/milestones/
    """
    def _update_attributes(self, mile):
        self._api = mile.get('url', '')
        #: Identifying number associated with milestone.
        self.number = mile.get('number')
        #: State of the milestone, e.g., open or closed.
        self.state = mile.get('state')
        #: Title of the milestone, e.g., 0.2.
        self.title = mile.get('title')
        #: Description of this milestone.
        self.description = mile.get('description')
        #: :class:`User <github3.users.User>` object representing the creator
        #: of the milestone.
        self.creator = None
        if mile.get('creator'):
            self.creator = User(mile.get('creator'), self)
        #: Number of issues associated with this milestone which are still
        #: open.
        self.open_issues = mile.get('open_issues')
        #: The number of closed issues associated with this milestone.
        self.closed_issues = mile.get('closed_issues')
        #: datetime object representing when the milestone was created.
        self.created_at = self._strptime(mile.get('created_at'))
        #: datetime representing when this milestone is due.
        self.due_on = self._strptime(mile.get('due_on'))
        #: datetime object representing when the milestone was updated.
        self.updated_at = self._strptime(mile.get('updated_at'))
        #: string representing the milestone's ID.
        self.id = mile.get('id')

    def _repr(self):
        return '<Milestone [{0}]>'.format(self)

    def __str__(self):
        return self.title

    @requires_auth
    def delete(self):
        """Delete this milestone.

        :returns: bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    def labels(self, number=-1, etag=None):
        r"""Iterate over the labels of every associated issue.

        .. versionchanged:: 0.9

            Add etag parameter.

        :param int number: (optional), number of labels to return. Default: -1
            returns all available labels.
        :param str etag: (optional), ETag header from a previous response
        :returns: generator of :class:`Label <github3.issues.label.Label>`\ s
        """
        url = self._build_url('labels', base_url=self._api)
        return self._iter(int(number), url, Label, etag=etag)

    @requires_auth
    def update(self, title=None, state=None, description=None, due_on=None):
        """Update this milestone.

        All parameters are optional, but it makes no sense to omit all of them
        at once.

        :param str title: (optional), new title of the milestone
        :param str state: (optional), ('open', 'closed')
        :param str description: (optional)
        :param str due_on: (optional), ISO 8601 time format:
            YYYY-MM-DDTHH:MM:SSZ
        :returns: bool
        """
        data = {'title': title, 'state': state,
                'description': description, 'due_on': due_on}
        self._remove_none(data)
        json = None

        if data:
            json = self._json(self._patch(self._api, data=dumps(data)), 200)
        if json:
            self._update_attributes(json)
            return True
        return False



# -*- coding: utf-8 -*-
"""A (tab-)completer for xonsh."""
import builtins
import collections.abc as cabc


class Completer(object):
    """This provides a list of optional completions for the xonsh shell."""

    def complete(self, prefix, line, begidx, endidx, ctx=None):
        """Complete the string, given a possible execution context.

        Parameters
        ----------
        prefix : str
            The string to match
        line : str
            The line that prefix appears on.
        begidx : int
            The index in line that prefix starts on.
        endidx : int
            The index in line that prefix ends on.
        ctx : Iterable of str (ie dict, set, etc), optional
            Names in the current execution context.

        Returns
        -------
        rtn : list of str
            Possible completions of prefix, sorted alphabetically.
        lprefix : int
            Length of the prefix to be replaced in the completion.
        """
        ctx = ctx or {}
        for func in builtins.__xonsh__.completers.values():
            try:
                out = func(prefix, line, begidx, endidx, ctx)
            except StopIteration:
                return set(), len(prefix)
            if isinstance(out, cabc.Sequence):
                res, lprefix = out
            else:
                res = out
                lprefix = len(prefix)
            if res is not None and len(res) != 0:

                def sortkey(s):
                    return s.lstrip(''''"''').lower()

                return tuple(sorted(res, key=sortkey)), lprefix
        return set(), lprefix



# -*- coding: utf-8 -*-
"""
github3.models
==============

This module provides the basic models used in github3.py

"""
from __future__ import unicode_literals

from json import dumps, loads
from requests.compat import urlparse, is_py2
from datetime import datetime
from logging import getLogger

from . import exceptions
from .decorators import requires_auth
from .null import NullObject
from .session import GitHubSession
from .utils import UTC

__timeformat__ = '%Y-%m-%dT%H:%M:%SZ'
__logs__ = getLogger(__package__)


class GitHubCore(object):

    """The base object for all objects that require a session.

    The :class:`GitHubCore <GitHubCore>` object provides some
    basic attributes and methods to other sub-classes that are very useful to
    have.
    """

    def __init__(self, json, session=None):
        if hasattr(session, 'session'):
            # i.e. session is actually a GitHubCore instance
            session = session.session
        elif session is None:
            session = GitHubSession()
        self.session = session

        # set a sane default
        self._github_url = 'https://api.github.com'

        if json is not None:
            self.etag = json.pop('ETag', None)
            self.last_modified = json.pop('Last-Modified', None)
            self._uniq = json.get('url', None)
        self._json_data = json
        self._update_attributes(json)

    def _update_attributes(self, json):
        pass

    def __getattr__(self, attribute):
        """Proxy access to stored JSON."""
        if attribute not in self._json_data:
            raise AttributeError(attribute)
        value = self._json_data.get(attribute, None)
        setattr(self, attribute, value)
        return value

    def as_dict(self):
        """Return the attributes for this object as a dictionary.

        This is equivalent to calling::

            json.loads(obj.as_json())

        :returns: this object's attributes serialized to a dictionary
        :rtype: dict
        """
        return self._json_data

    def as_json(self):
        """Return the json data for this object.

        This is equivalent to calling::

            json.dumps(obj.as_dict())

        :returns: this object's attributes as a JSON string
        :rtype: str
        """
        return dumps(self._json_data)

    def _strptime(self, time_str):
        """Convert an ISO 8601 formatted string to a datetime object.

        We assume that the ISO 8601 formatted string is in UTC and we create
        the datetime object so that it is timezone-aware.

        :param str time_str: ISO 8601 formatted string
        :returns: timezone-aware datetime object
        :rtype: datetime or None
        """
        if time_str:
            # Parse UTC string into naive datetime, then add timezone
            dt = datetime.strptime(time_str, __timeformat__)
            return dt.replace(tzinfo=UTC())
        return None

    def __repr__(self):
        repr_string = self._repr()
        if is_py2:
            return repr_string.encode('utf-8')
        return repr_string

    @classmethod
    def from_dict(cls, json_dict):
        """Return an instance of this class formed from ``json_dict``."""
        return cls(json_dict)

    @classmethod
    def from_json(cls, json):
        """Return an instance of this class formed from ``json``."""
        return cls(loads(json))

    def __eq__(self, other):
        return self._uniq == other._uniq

    def __ne__(self, other):
        return self._uniq != other._uniq

    def __hash__(self):
        return hash(self._uniq)

    def _repr(self):
        return '<github3-core at 0x{0:x}>'.format(id(self))

    @staticmethod
    def _remove_none(data):
        if not data:
            return
        for (k, v) in list(data.items()):
            if v is None:
                del(data[k])

    def _instance_or_null(self, instance_class, json):
        if json is None:
            return NullObject(instance_class.__name__)
        if not isinstance(json, dict):
            return exceptions.UnprocessableResponseBody(
                "GitHub's API returned a body that could not be handled", json
            )
        try:
            return instance_class(json, self)
        except TypeError:  # instance_class is not a subclass of GitHubCore
            return instance_class(json)

    def _json(self, response, status_code, include_cache_info=True):
        ret = None
        if self._boolean(response, status_code, 404) and response.content:
            __logs__.info('Attempting to get JSON information from a Response '
                          'with status code %d expecting %d',
                          response.status_code, status_code)
            ret = response.json()
            headers = response.headers
            if (include_cache_info and
                    (headers.get('Last-Modified') or headers.get('ETag')) and
                    isinstance(ret, dict)):
                ret['Last-Modified'] = response.headers.get(
                    'Last-Modified', ''
                )
                ret['ETag'] = response.headers.get('ETag', '')
        __logs__.info('JSON was %sreturned', 'not ' if ret is None else '')
        return ret

    def _boolean(self, response, true_code, false_code):
        if response is not None:
            status_code = response.status_code
            if status_code == true_code:
                return True
            if status_code != false_code and status_code >= 400:
                raise exceptions.error_for(response)
        return False

    def _delete(self, url, **kwargs):
        __logs__.debug('DELETE %s with %s', url, kwargs)
        return self.session.delete(url, **kwargs)

    def _get(self, url, **kwargs):
        __logs__.debug('GET %s with %s', url, kwargs)
        return self.session.get(url, **kwargs)

    def _patch(self, url, **kwargs):
        __logs__.debug('PATCH %s with %s', url, kwargs)
        return self.session.patch(url, **kwargs)

    def _post(self, url, data=None, json=True, **kwargs):
        if json:
            data = dumps(data) if data is not None else data
        __logs__.debug('POST %s with %s, %s', url, data, kwargs)
        return self.session.post(url, data, **kwargs)

    def _put(self, url, **kwargs):
        __logs__.debug('PUT %s with %s', url, kwargs)
        return self.session.put(url, **kwargs)

    def _build_url(self, *args, **kwargs):
        """Builds a new API url from scratch."""
        return self.session.build_url(*args, **kwargs)

    @property
    def _api(self):
        return "{0.scheme}://{0.netloc}{0.path}".format(self._uri)

    @_api.setter
    def _api(self, uri):
        self._uri = urlparse(uri)
        self.url = uri

    def _iter(self, count, url, cls, params=None, etag=None, headers=None):
        """Generic iterator for this project.

        :param int count: How many items to return.
        :param int url: First URL to start with
        :param class cls: cls to return an object of
        :param params dict: (optional) Parameters for the request
        :param str etag: (optional), ETag from the last call
        :param dict headers: (optional) HTTP Headers for the request
        :returns: A lazy iterator over the pagianted resource
        :rtype: :class:`GitHubIterator <github3.structs.GitHubIterator>`
        """
        from .structs import GitHubIterator
        return GitHubIterator(count, url, cls, self, params, etag, headers)

    @property
    def ratelimit_remaining(self):
        """Number of requests before GitHub imposes a ratelimit.

        :returns: int
        """
        json = self._json(self._get(self._github_url + '/rate_limit'), 200)
        core = json.get('resources', {}).get('core', {})
        self._remaining = core.get('remaining', 0)
        return self._remaining

    def refresh(self, conditional=False):
        """Re-retrieve the information for this object.

        The reasoning for the return value is the following example: ::

            repos = [r.refresh() for r in g.repositories_by('kennethreitz')]

        Without the return value, that would be an array of ``None``'s and you
        would otherwise have to do: ::

            repos = [r for i in g.repositories_by('kennethreitz')]
            [r.refresh() for r in repos]

        Which is really an anti-pattern.

        .. versionchanged:: 0.5

        .. _Conditional Requests:
            http://developer.github.com/v3/#conditional-requests

        :param bool conditional: If True, then we will search for a stored
            header ('Last-Modified', or 'ETag') on the object and send that
            as described in the `Conditional Requests`_ section of the docs
        :returns: self
        """
        headers = {}
        if conditional:
            if self.last_modified:
                headers['If-Modified-Since'] = self.last_modified
            elif self.etag:
                headers['If-None-Match'] = self.etag

        headers = headers or None
        json = self._json(self._get(self._api, headers=headers), 200)
        if json is not None:
            self._update_attributes(json)
        return self


class BaseComment(GitHubCore):

    """A basic class for Gist, Issue and Pull Request Comments."""

    def _update_attributes(self, comment):
        #: Unique ID of the comment.
        self.id = comment.get('id')
        #: Body of the comment. (As written by the commenter)
        self.body = comment.get('body')
        #: Body of the comment formatted as plain-text. (Stripped of markdown,
        #: etc.)
        self.body_text = comment.get('body_text')
        #: Body of the comment formatted as html.
        self.body_html = comment.get('body_html')
        #: datetime object representing when the comment was created.
        self.created_at = self._strptime(comment.get('created_at'))
        #: datetime object representing when the comment was updated.
        self.updated_at = self._strptime(comment.get('updated_at'))

        self._api = comment.get('url', '')
        self.links = comment.get('_links')
        #: The url of this comment at GitHub
        self.html_url = ''
        #: The url of the pull request, if it exists
        self.pull_request_url = ''
        if self.links:
            self.html_url = self.links.get('html')
            self.pull_request_url = self.links.get('pull_request')

    @requires_auth
    def delete(self):
        """Delete this comment.

        :returns: bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def edit(self, body):
        """Edit this comment.

        :param str body: (required), new body of the comment, Markdown
            formatted
        :returns: bool
        """
        if body:
            json = self._json(self._patch(self._api,
                              data=dumps({'body': body})), 200)
            if json:
                self._update_attributes(json)
                return True
        return False


class BaseCommit(GitHubCore):

    """This abstracts a lot of the common attributes for commit-like objects.

    The :class:`BaseCommit <BaseCommit>` object serves as the base for
    the various types of commit objects returned by the API.
    """

    def _update_attributes(self, commit):
        self._api = commit.get('url', '')
        #: SHA of this commit.
        self.sha = commit.get('sha')
        #: Commit message
        self.message = commit.get('message')
        #: List of parents to this commit.
        self.parents = commit.get('parents', [])
        #: URL to view the commit on GitHub
        self.html_url = commit.get('html_url', '')
        if not self.sha:
            i = self._api.rfind('/')
            self.sha = self._api[i + 1:]

        self._uniq = self.sha


class BaseAccount(GitHubCore):

    """This class holds the commonalities of Organizations and Users.

    The :class:`BaseAccount <BaseAccount>` object is used to do the
    heavy lifting for :class:`Organization <github3.orgs.Organization>` and
    :class:`User <github3.users.User>` objects.
    """

    def _update_attributes(self, acct):
        #: Tells you what type of account this is
        self.type = None
        if acct.get('type'):
            self.type = acct.get('type')
        self._api = acct.get('url', '')

        #: URL of the avatar at gravatar
        self.avatar_url = acct.get('avatar_url', '')
        #: URL of the blog
        self.blog = acct.get('blog', '')
        #: Name of the company
        self.company = acct.get('company', '')

        #: datetime object representing the date the account was created
        self.created_at = self._strptime(acct.get('created_at'))

        #: E-mail address of the user/org
        self.email = acct.get('email')

        # The number of people following this acct
        #: Number of followers
        self.followers_count = acct.get('followers', 0)

        # The number of people this acct follows
        #: Number of people the user is following
        self.following_count = acct.get('following', 0)

        #: Unique ID of the account
        self.id = acct.get('id', 0)
        #: Location of the user/org
        self.location = acct.get('location', '')
        #: User name of the user/organization
        self.login = acct.get('login', '')

        # e.g. first_name last_name
        #: Real name of the user/org
        self.name = acct.get('name') or ''
        self.name = self.name

        # The number of public_repos
        #: Number of public repos owned by the user/org
        self.public_repos_count = acct.get('public_repos', 0)

        # e.g. https://github.com/self._login
        #: URL of the user/org's profile
        self.html_url = acct.get('html_url', '')

        #: Markdown formatted biography
        self.bio = acct.get('bio')

    def _repr(self):
        return '<{s.type} [{s.login}:{s.name}]>'.format(s=self)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from json import dumps
from ..models import GitHubCore
from .commit import RepoCommit


class Branch(GitHubCore):
    """The :class:`Branch <Branch>` object. It holds the information GitHub
    returns about a branch on a
    :class:`Repository <github3.repos.repo.Repository>`.
    """

    # The Accept header will likely be removable once the feature is out of
    # preview mode. See: http://git.io/v4O1e
    PREVIEW_HEADERS = {'Accept': 'application/vnd.github.loki-preview+json'}

    def _update_attributes(self, branch):
        #: Name of the branch.
        self.name = branch.get('name')
        #: Returns the branch's
        #: :class:`RepoCommit <github3.repos.commit.RepoCommit>` or ``None``.
        self.commit = branch.get('commit')
        if self.commit:
            self.commit = RepoCommit(self.commit, self)
        #: Returns '_links' attribute.
        self.links = branch.get('_links', {})
        #: Provides the branch's protection status.
        self.protection = branch.get('protection')

        if 'self' in self.links:
            self._api = self.links['self']
        else:  # Branches obtained via `repo.branches` don't have links.
            base = self.commit.url.split('/commit', 1)[0]
            self._api = self._build_url('branches', self.name, base_url=base)

    def _repr(self):
        return '<Repository Branch [{0}]>'.format(self.name)

    def latest_sha(self, differs_from=''):
        """Check if SHA-1 is the same as remote branch

        See: https://git.io/vaqIw

        :param str differs_from: (optional), sha to compare against
        :returns: string of the SHA or None
        """
        # If-None-Match returns 200 instead of 304 value does not have quotes
        headers = {
            'Accept': 'application/vnd.github.chitauri-preview+sha',
            'If-None-Match': '"{0}"'.format(differs_from)
        }
        base = self._api.split('/branches', 1)[0]
        url = self._build_url('commits', self.name, base_url=base)
        resp = self._get(url, headers=headers)
        if self._boolean(resp, 200, 304):
            return resp.content
        return None

    def protect(self, enforcement=None, status_checks=None):
        """Enable force push protection and configure status check enforcement.

        See: http://git.io/v4Gvu

        :param str enforcement: (optional), Specifies the enforcement level of
            the status checks. Must be one of 'off', 'non_admins', or
            'everyone'. Use `None` or omit to use the already associated value.
        :param list status_checks: (optional), An list of strings naming
            status checks that must pass before merging. Use `None` or omit to
            use the already associated value.
        """
        previous_values = self.protection['required_status_checks']
        if enforcement is None:
            enforcement = previous_values['enforcement_level']
        if status_checks is None:
            status_checks = previous_values['contexts']

        edit = {'protection': {'enabled': True, 'required_status_checks': {
            'enforcement_level': enforcement, 'contexts': status_checks}}}
        json = self._json(self._patch(self._api, data=dumps(edit),
                                      headers=self.PREVIEW_HEADERS), 200)
        self._update_attributes(json)
        return True

    def unprotect(self):
        """Disable force push protection on this branch."""
        edit = {'protection': {'enabled': False}}
        json = self._json(self._patch(self._api, data=dumps(edit),
                                      headers=self.PREVIEW_HEADERS), 200)
        self._update_attributes(json)
        return True



"""Xonsh completer tools."""
import builtins
import textwrap


def _filter_normal(s, x):
    return s.startswith(x)


def _filter_ignorecase(s, x):
    return s.lower().startswith(x.lower())


def get_filter_function():
    """
    Return an appropriate filtering function for completions, given the valid
    of $CASE_SENSITIVE_COMPLETIONS
    """
    csc = builtins.__xonsh__.env.get("CASE_SENSITIVE_COMPLETIONS")
    if csc:
        return _filter_normal
    else:
        return _filter_ignorecase


def justify(s, max_length, left_pad=0):
    """
    Re-wrap the string s so that each line is no more than max_length
    characters long, padding all lines but the first on the left with the
    string left_pad.
    """
    txt = textwrap.wrap(s, width=max_length, subsequent_indent=" " * left_pad)
    return "\n".join(txt)



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import print_function
from __future__ import division

import calendar
from datetime import datetime
import pytz
import time


def pretty_date_time(date_time):
    """Prints a pretty datetime similar to what's seen on Hacker News.

    Gets a datetime object or a int() Epoch timestamp and return a
    pretty string like 'an hour ago', 'Yesterday', '3 months ago',
    'just now', etc.

    Adapted from: http://stackoverflow.com/questions/1551382/user-friendly-time-format-in-python  # NOQA

    Args:
        * date_time: An instance of datetime.

    Returns:
        A string that represents the pretty datetime.
    """
    now = datetime.now(pytz.utc)
    if type(date_time) is int:
        diff = now - datetime.fromtimestamp(date_time)
    elif type(date_time) is time.struct_time:
        date_time = datetime.fromtimestamp(calendar.timegm(date_time),
                                           tz=pytz.utc)
        diff = now - date_time
    elif isinstance(date_time, datetime):
        diff = now - date_time
    elif not date_time:
        diff = now - now
    second_diff = diff.seconds
    day_diff = diff.days
    if day_diff < 0:
        return ''
    if day_diff == 0:
        if second_diff < 10:
            return "just now"
        if second_diff < 60:
            return str(second_diff) + " seconds ago"
        if second_diff < 120:
            return "1 minute ago"
        if second_diff < 3600:
            return str(second_diff // 60) + " minutes ago"
        if second_diff < 7200:
            return "1 hour ago"
        if second_diff < 86400:
            return str(second_diff // 3600) + " hours ago"
    if day_diff == 1:
        return "Yesterday"
    if day_diff < 7:
        return str(day_diff) + " days ago"
    if day_diff < 31:
        return str(day_diff // 7) + " week(s) ago"
    if day_diff < 365:
        return str(day_diff // 30) + " month(s) ago"
    return str(day_diff // 365) + " year(s) ago"



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

formatted_user_feed = u'\x1b[35m  1.   \x1b[0mLeonardoGitHub \x1b[0m\x1b[32mforked \x1b[0m\x1b[36mdonnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  2.   \x1b[0mnavroopsingh \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/dev-setup \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  3.   \x1b[0mprabaharanr \x1b[0m\x1b[32mforked \x1b[0m\x1b[36mdonnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  4.   \x1b[0mwdavis12 \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  5.   \x1b[0mvonec \x1b[0m\x1b[32mcommented on issue\x1b[0m\x1b[36m awslabs/aws-shell#115 \x1b[0m\x1b[33m(just now)\x1b[0m\n         \x1b[36m\x1b[0mFixed     \x1b[0m\n\x1b[35m  6.   \x1b[0mbenc \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/dev-setup \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  7.   \x1b[0mryusun \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  8.   \x1b[0mchirag200666 \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  9.   \x1b[0mravi-kothari \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/interactive-coding-challenges \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  10.  \x1b[0mravi-kothari \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  11.  \x1b[0mmarnovo \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  12.  \x1b[0mandersy005 \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/interactive-coding-challenges \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  13.  \x1b[0mandersy005 \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/viz \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  14.  \x1b[0mt9md \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/interactive-coding-challenges \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  15.  \x1b[0msilivizzle \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/dev-setup \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  16.  \x1b[0mahna \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  17.  \x1b[0mziyasal \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/saws \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  18.  \x1b[0mchefablz \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  19.  \x1b[0mkevinkn \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/dev-setup \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  20.  \x1b[0mfrrp \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/interactive-coding-challenges \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  21.  \x1b[0mfrrp \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  22.  \x1b[0mLiberTang0 \x1b[0m\x1b[32mforked \x1b[0m\x1b[36mdonnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  23.  \x1b[0mPhilippe-Caparroy \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  24.  \x1b[0mPhilippe-Caparroy \x1b[0m\x1b[32mforked \x1b[0m\x1b[36mdonnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  25.  \x1b[0mtiendan \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  26.  \x1b[0mSkyfold \x1b[0m\x1b[32mclosed issue\x1b[0m\x1b[36m awslabs/aws-shell#115 \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  27.  \x1b[0mSkyfold \x1b[0m\x1b[32mcommented on issue\x1b[0m\x1b[36m awslabs/aws-shell#115 \x1b[0m\x1b[33m(just now)\x1b[0m\n         \x1b[36m\x1b[0mFixed     \x1b[0m\n\x1b[35m  28.  \x1b[0msouadah \x1b[0m\x1b[32mforked \x1b[0m\x1b[36mdonnemartin/data-science-ipython-notebooks \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[35m  29.  \x1b[0mjoshuadsilva \x1b[0m\x1b[32mcommented on issue\x1b[0m\x1b[36m awslabs/aws-shell#115 \x1b[0m\x1b[33m(just now)\x1b[0m\n         \x1b[36m\x1b[0mWorks for me too img class="emoji" title=":+1:"\n         alt=":+1:" src="https://assets-\n         cdn.github.com/images/icons/emoji/unicode/1f44d.png" height="20"\n         width="20" align="absmiddle"\x1b[0m\n\x1b[35m  30.  \x1b[0mwilsantos99 \x1b[0m\x1b[32mstarred\x1b[0m\x1b[36m donnemartin/awesome-aws \x1b[0m\x1b[33m(just now)\x1b[0m\n\x1b[0m'



import os
import re
import pickle
import builtins
import subprocess

import xonsh.lazyasd as xl

from xonsh.completers.tools import get_filter_function

OPTIONS = None
OPTIONS_PATH = None


@xl.lazyobject
def SCRAPE_RE():
    return re.compile(r"^(?:\s*(?:-\w|--[a-z0-9-]+)[\s,])+", re.M)


@xl.lazyobject
def INNER_OPTIONS_RE():
    return re.compile(r"-\w|--[a-z0-9-]+")


def complete_from_man(prefix, line, start, end, ctx):
    """
    Completes an option name, based on the contents of the associated man
    page.
    """
    global OPTIONS, OPTIONS_PATH
    if OPTIONS is None:
        datadir = builtins.__xonsh__.env["XONSH_DATA_DIR"]
        OPTIONS_PATH = os.path.join(datadir, "man_completions_cache")
        try:
            with open(OPTIONS_PATH, "rb") as f:
                OPTIONS = pickle.load(f)
        except Exception:
            OPTIONS = {}
    if not prefix.startswith("-"):
        return set()
    cmd = line.split()[0]
    if cmd not in OPTIONS:
        try:
            manpage = subprocess.Popen(
                ["man", cmd], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
            )
            # This is a trick to get rid of reverse line feeds
            text = subprocess.check_output(["col", "-b"], stdin=manpage.stdout)
            text = text.decode("utf-8")
            scraped_text = " ".join(SCRAPE_RE.findall(text))
            matches = INNER_OPTIONS_RE.findall(scraped_text)
            OPTIONS[cmd] = matches
            with open(OPTIONS_PATH, "wb") as f:
                pickle.dump(OPTIONS, f)
        except Exception:
            return set()
    return {s for s in OPTIONS[cmd] if get_filter_function()(s, prefix)}



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

formatted_gitignores = u'\x1b[35m  1.   \x1b[0mActionscript \x1b[0m\n\x1b[35m  2.   \x1b[0mAda \x1b[0m\n\x1b[35m  3.   \x1b[0mAgda \x1b[0m\n\x1b[35m  4.   \x1b[0mAndroid \x1b[0m\n\x1b[35m  5.   \x1b[0mAppEngine \x1b[0m\n\x1b[0m'
formatted_gitignores_tip = u'  Run the following command to view or download a .gitignore file:\n    View:     gh gitignore Python\n    Download: gh gitignore Python > .gitignore\n'



import builtins
import collections

import xonsh.lazyasd as xl

from xonsh.completers.tools import justify


VALID_ACTIONS = xl.LazyObject(
    lambda: frozenset({"add", "remove", "list"}), globals(), "VALID_ACTIONS"
)


def _add_one_completer(name, func, loc="end"):
    new = collections.OrderedDict()
    if loc == "start":
        new[name] = func
        for (k, v) in builtins.__xonsh__.completers.items():
            new[k] = v
    elif loc == "end":
        for (k, v) in builtins.__xonsh__.completers.items():
            new[k] = v
        new[name] = func
    else:
        direction, rel = loc[0], loc[1:]
        found = False
        for (k, v) in builtins.__xonsh__.completers.items():
            if rel == k and direction == "<":
                new[name] = func
                found = True
            new[k] = v
            if rel == k and direction == ">":
                new[name] = func
                found = True
        if not found:
            new[name] = func
    builtins.__xonsh__.completers.clear()
    builtins.__xonsh__.completers.update(new)


def _list_completers(args, stdin=None, stack=None):
    o = "Registered Completer Functions: \n"
    _comp = builtins.__xonsh__.completers
    ml = max((len(i) for i in _comp), default=0)
    _strs = []
    for c in _comp:
        if _comp[c].__doc__ is None:
            doc = "No description provided"
        else:
            doc = " ".join(_comp[c].__doc__.split())
        doc = justify(doc, 80, ml + 3)
        _strs.append("{: >{}} : {}".format(c, ml, doc))
    return o + "\n".join(_strs) + "\n"


def _remove_completer(args, stdin=None, stack=None):
    err = None
    if len(args) != 1:
        err = "completer remove takes exactly 1 argument."
    else:
        name = args[0]
        if name not in builtins.__xonsh__.completers:
            err = ("The name %s is not a registered " "completer function.") % name
    if err is None:
        del builtins.__xonsh__.completers[name]
        return
    else:
        return None, err + "\n", 1


def _register_completer(args, stdin=None, stack=None):
    err = None
    if len(args) not in {2, 3}:
        err = (
            "completer add takes either 2 or 3 arguments.\n"
            "For help, run:  completer help add"
        )
    else:
        name = args[0]
        func_name = args[1]
        if name in builtins.__xonsh__.completers:
            err = ("The name %s is already a registered " "completer function.") % name
        else:
            if func_name in builtins.__xonsh__.ctx:
                func = builtins.__xonsh__.ctx[func_name]
                if not callable(func):
                    err = "%s is not callable" % func_name
            else:
                for frame_info in stack:
                    frame = frame_info[0]
                    if func_name in frame.f_locals:
                        func = frame.f_locals[func_name]
                        break
                    elif func_name in frame.f_globals:
                        func = frame.f_globals[func_name]
                        break
                else:
                    err = "No such function: %s" % func_name
    if err is None:
        position = "start" if len(args) == 2 else args[2]
        _add_one_completer(name, func, position)
    else:
        return None, err + "\n", 1


def completer_alias(args, stdin=None, stdout=None, stderr=None, spec=None, stack=None):
    err = None
    if len(args) == 0 or args[0] not in (VALID_ACTIONS | {"help"}):
        err = (
            "Please specify an action.  Valid actions are: "
            '"add", "remove", "list", or "help".'
        )
    elif args[0] == "help":
        if len(args) == 1 or args[1] not in VALID_ACTIONS:
            return (
                "Valid actions are: add, remove, list.  For help with a "
                "specific action, run: completer help ACTION\n"
            )
        elif args[1] == "add":
            return COMPLETER_ADD_HELP_STR
        elif args[1] == "remove":
            return COMPLETER_REMOVE_HELP_STR
        elif args[1] == "list":
            return COMPLETER_LIST_HELP_STR

    if err is not None:
        return None, err + "\n", 1

    if args[0] == "add":
        func = _register_completer
    elif args[0] == "remove":
        func = _remove_completer
    elif args[0] == "list":
        func = _list_completers
    return func(args[1:], stdin=stdin, stack=stack)


COMPLETER_LIST_HELP_STR = """completer list: ordered list the active completers

Usage:
    completer remove
"""

COMPLETER_REMOVE_HELP_STR = """completer remove: removes a completer from xonsh

Usage:
    completer remove NAME

NAME is a unique name of a completer (run "completer list" to see the current
     completers in order)
"""

COMPLETER_ADD_HELP_STR = """completer add: adds a new completer to xonsh

Usage:
    completer add NAME FUNC [POS]

NAME is a unique name to use in the listing (run "completer list" to see the
     current completers in order)

FUNC is the name of a completer function to use.  This should be a function
     of the following arguments, and should return a set of valid completions
     for the given prefix.  If this completer should not be used in a given
     context, it should return an empty set or None.

     Arguments to FUNC:
       * prefix: the string to be matched
       * line: a string representing the whole current line, for context
       * begidx: the index at which prefix starts in line
       * endidx: the index at which prefix ends in line
       * ctx: the current Python environment

     If the completer expands the prefix in any way, it should return a tuple
     of two elements: the first should be the set of completions, and the
     second should be the length of the modified prefix (for an example, see
     xonsh.completers.path.complete_path).

POS (optional) is a position into the list of completers at which the new
     completer should be added.  It can be one of the following values:
       * "start" indicates that the completer should be added to the start of
                 the list of completers (it should be run before all others)
       * "end" indicates that the completer should be added to the end of the
               list of completers (it should be run after all others)
       * ">KEY", where KEY is a pre-existing name, indicates that this should
                 be added after the completer named KEY
       * "<KEY", where KEY is a pre-existing name, indicates that this should
                 be added before the completer named KEY

     If POS is not provided, the default value is "start"
"""



# -*- coding: utf-8 -*-
"""Tools to replay xonsh history files."""
import json
import time
import builtins
import collections.abc as cabc

from xonsh.tools import swap
from xonsh.lazyjson import LazyJSON
from xonsh.environ import Env
import xonsh.history.main as xhm


DEFAULT_MERGE_ENVS = ("replay", "native")


class Replayer(object):
    """Replays a xonsh history file."""

    def __init__(self, f, reopen=True):
        """
        Parameters
        ----------
        f : file handle or str
            Path to xonsh history file.
        reopen : bool, optional
            Whether new file handle should be opened for each load, passed directly into
            LazyJSON class.
        """
        self._lj = LazyJSON(f, reopen=reopen)

    def __del__(self):
        self._lj.close()

    def replay(self, merge_envs=DEFAULT_MERGE_ENVS, target=None):
        """Replays the history specified, returns the history object where the code
        was executed.

        Parameters
        ----------
        merge_env : tuple of str or Mappings, optional
            Describes how to merge the environments, in order of increasing precedence.
            Available strings are 'replay' and 'native'. The 'replay' env comes from the
            history file that we are replaying. The 'native' env comes from what this
            instance of xonsh was started up with. Instead of a string, a dict or other
            mapping may be passed in as well. Defaults to ('replay', 'native').
        target : str, optional
            Path to new history file.
        """
        shell = builtins.__xonsh__.shell
        re_env = self._lj["env"].load()
        new_env = self._merge_envs(merge_envs, re_env)
        new_hist = xhm.construct_history(
            env=new_env.detype(),
            locked=True,
            ts=[time.time(), None],
            gc=False,
            filename=target,
        )
        with swap(builtins.__xonsh__, "env", new_env), swap(
            builtins.__xonsh__, "history", new_hist
        ):
            for cmd in self._lj["cmds"]:
                inp = cmd["inp"]
                shell.default(inp)
                if builtins.__xonsh__.exit:  # prevent premature exit
                    builtins.__xonsh__.exit = False
        new_hist.flush(at_exit=True)
        return new_hist

    def _merge_envs(self, merge_envs, re_env):
        new_env = {}
        for e in merge_envs:
            if e == "replay":
                new_env.update(re_env)
            elif e == "native":
                new_env.update(builtins.__xonsh__.env)
            elif isinstance(e, cabc.Mapping):
                new_env.update(e)
            else:
                raise TypeError("Type of env not understood: {0!r}".format(e))
        new_env = Env(**new_env)
        return new_env


_REPLAY_PARSER = None


def replay_create_parser(p=None):
    global _REPLAY_PARSER
    p_was_none = p is None
    if _REPLAY_PARSER is not None and p_was_none:
        return _REPLAY_PARSER
    if p_was_none:
        from argparse import ArgumentParser

        p = ArgumentParser("replay", description="replays a xonsh history file")
    p.add_argument(
        "--merge-envs",
        dest="merge_envs",
        default=DEFAULT_MERGE_ENVS,
        nargs="+",
        help="Describes how to merge the environments, in order of "
        "increasing precedence. Available strings are 'replay' and "
        "'native'. The 'replay' env comes from the history file that we "
        "are replaying. The 'native' env comes from what this instance "
        "of xonsh was started up with. One or more of these options may "
        "be passed in. Defaults to '--merge-envs replay native'.",
    )
    p.add_argument(
        "--json",
        dest="json",
        default=False,
        action="store_true",
        help="print history info in JSON format",
    )
    p.add_argument(
        "-o", "--target", dest="target", default=None, help="path to new history file"
    )
    p.add_argument("path", help="path to replay history file")
    if p_was_none:
        _REPLAY_PARSER = p
    return p


def replay_main_action(h, ns, stdout=None, stderr=None):
    replayer = Replayer(ns.path)
    hist = replayer.replay(merge_envs=ns.merge_envs, target=ns.target)
    print("----------------------------------------------------------------")
    print("Just replayed history, new history has the following information")
    print("----------------------------------------------------------------")
    data = hist.info()
    if ns.json:
        s = json.dumps(data)
        print(s, file=stdout)
    else:
        lines = ["{0}: {1}".format(k, v) for k, v in data.items()]
        print("\n".join(lines), file=stdout)


def replay_main(args, stdin=None):
    """Acts as main function for replaying a xonsh history file."""
    parser = replay_create_parser()
    ns = parser.parse_args(args)
    replay_main_action(ns)



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import print_function
from __future__ import division

import click
import time


def timeit(method):
    """From: https://www.andreas-jung.com/contents/a-python-decorator-for-measuring-the-execution-time-of-methods  # NOQA
    """
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        message = '%r (%r, %r) %2.2f sec' % (method.__name__, args, kw, te-ts)
        click.secho(message + '\n', fg='red')
        return result
    return timed



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import unicode_literals
from __future__ import print_function

import mock

from compat import unittest

from gitsome.github import GitHub
from tests.mock_feed_parser import MockFeedParser
from tests.mock_github_api import MockGitHubApi
from tests.mock_pretty_date_time import pretty_date_time
from tests.data.email import formatted_emails
from tests.data.emoji import formatted_emojis
from tests.data.events import formatted_events
from tests.data.user import formatted_org, formatted_user, formatted_users
from tests.data.gitignores import formatted_gitignores, formatted_gitignores_tip
from tests.data.issue import formatted_issues, formatted_pull_requests
from tests.data.license import formatted_licenses, formatted_licenses_tip
from tests.data.thread import formatted_threads
from tests.data.trends import formatted_trends
from tests.data.user_feed import formatted_user_feed


class GitHubTest(unittest.TestCase):

    def setUp(self):
        self.github = GitHub()
        self.github.config.api = MockGitHubApi()
        self.github.formatter.pretty_dt = pretty_date_time
        self.github.trend_parser = MockFeedParser()

    def test_avatar_no_pil(self):
        avatar_text = self.github.avatar(
            'https://avatars.githubusercontent.com/u/583231?v=3', False)
        assert avatar_text == 'PIL not found.\n'

    @mock.patch('gitsome.github.click.secho')
    def test_create_comment(self, mock_click_secho):
        self.github.create_comment('user1/repo1/1', 'text')
        mock_click_secho.assert_called_with(
            'Created comment: text',
            fg=self.github.config.clr_message)

    @mock.patch('gitsome.github.click.secho')
    def test_create_comment_invalid_args(self, mock_click_secho):
        self.github.create_comment('invalid/repo1/1', 'text')
        mock_click_secho.assert_called_with(
            'Error creating comment',
            fg=self.github.config.clr_error)
        self.github.create_comment('user1/repo1/foo', 'text')
        mock_click_secho.assert_called_with(
            'Expected argument: user/repo/# and option -t "comment".',
            fg=self.github.config.clr_error)

    @mock.patch('gitsome.github.click.secho')
    def test_create_issue(self, mock_click_secho):
        self.github.create_issue('user1/repo1', 'title', 'desc')
        mock_click_secho.assert_called_with(
            'Created issue: title\ndesc',
            fg=self.github.config.clr_message)

    @mock.patch('gitsome.github.click.secho')
    def test_create_issue_no_desc(self, mock_click_secho):
        self.github.create_issue('user1/repo1', 'title', issue_desc=None)
        mock_click_secho.assert_called_with(
            'Created issue: title\n',
            fg=self.github.config.clr_message)

    @mock.patch('gitsome.github.click.secho')
    def test_create_issue_invalid_args(self, mock_click_secho):
        self.github.create_issue('invalid/repo1', 'title', 'desc')
        mock_click_secho.assert_called_with(
            'Error creating issue.',
            fg=self.github.config.clr_error)
        self.github.create_issue('u', 'title', 'desc')
        mock_click_secho.assert_called_with(
            'Expected argument: user/repo and option -t "title".',
            fg=self.github.config.clr_error)

    @mock.patch('gitsome.github.click.secho')
    def test_create_repo(self, mock_click_secho):
        self.github.create_repo('name', 'desc', True)
        mock_click_secho.assert_called_with(
            'Created repo: name\ndesc',
            fg=self.github.config.clr_message)

    @mock.patch('gitsome.github.click.secho')
    def test_create_repo_no_desc(self, mock_click_secho):
        self.github.create_repo('name', repo_desc=None)
        mock_click_secho.assert_called_with(
            'Created repo: name\n',
            fg=self.github.config.clr_message)

    @mock.patch('gitsome.github.click.secho')
    def test_create_repo_invalid_args(self, mock_click_secho):
        self.github.create_repo('repo1', 'desc', True)
        mock_click_secho.assert_called_with(
            'Error creating repo: foobar',
            fg=self.github.config.clr_error)

    @mock.patch('gitsome.github.click.secho')
    def test_emails(self, mock_click_secho):
        self.github.emails()
        mock_click_secho.assert_called_with(formatted_emails)

    @mock.patch('gitsome.github.click.secho')
    @mock.patch('gitsome.config.Config.prompt_news_feed')
    def test_feed_config(self, mock_config_prompt_news_feed, mock_click_secho):
        self.github.feed()
        mock_config_prompt_news_feed.assert_called_with()

    @mock.patch('gitsome.github.click.secho')
    def test_feed(self, mock_click_secho):
        self.github.config.user_feed = 'user_feed'
        self.github.feed()
        mock_click_secho.assert_called_with(formatted_user_feed)

    @mock.patch('gitsome.github.click.secho')
    @mock.patch('gitsome.config.Config')
    def test_feed_user(self, mock_config, mock_click_secho):
        self.github.feed('user1')
        mock_click_secho.assert_called_with(formatted_events)

    @mock.patch('gitsome.github.click.secho')
    def test_emojis(self, mock_click_secho):
        self.github.emojis()
        mock_click_secho.assert_called_with(formatted_emojis)

    @mock.patch('gitsome.github.click.secho')
    def test_followers(self, mock_click_secho):
        self.github.followers('foo')
        mock_click_secho.assert_called_with(formatted_users)

    @mock.patch('gitsome.github.click.secho')
    def test_following(self, mock_click_secho):
        self.github.following('foo')
        mock_click_secho.assert_called_with(formatted_users)

    @mock.patch('gitsome.github.click.secho')
    def test_gitignore_template(self, mock_click_secho):
        self.github.gitignore_template('valid_language')
        mock_click_secho.assert_called_with(
            'template',
            fg=self.github.config.clr_message)

    @mock.patch('gitsome.github.click.secho')
    def test_gitignore_template_invalid(self, mock_click_secho):
        self.github.gitignore_template('invalid_language')
        mock_click_secho.assert_called_with(
            ('Invalid case-sensitive template requested, run the '
             'following command to see available templates:\n'
             '    gh gitignore-templates'),
            fg=self.github.config.clr_error)

    @mock.patch('gitsome.github.click.secho')
    def test_gitignore_templates(self, mock_click_secho):
        self.github.gitignore_templates()
        mock_click_secho.assert_any_call(formatted_gitignores)
        mock_click_secho.assert_any_call(formatted_gitignores_tip,
                                         fg=self.github.config.clr_message)

    @mock.patch('gitsome.web_viewer.WebViewer.view_url')
    def test_issue(self, mock_view_url):
        self.github.issue('user1/repo1/1')
        mock_view_url.assert_called_with(
            'https://github.com/user1/repo1/issues/1')

    @mock.patch('gitsome.github.click.secho')
    def test_issue_invalid_args(self, mock_click_secho):
        self.github.issue('user1/repo1/foo')
        mock_click_secho.assert_called_with(
            'Expected argument: user/repo/#.',
            fg=self.github.config.clr_error)

    @mock.patch('gitsome.github.click.secho')
    def test_issues_setup(self, mock_click_secho):
        self.github.issues_setup()
        mock_click_secho.assert_called_with(formatted_issues)

    @mock.patch('gitsome.github.click.secho')
    def test_license(self, mock_click_secho):
        self.github.license('valid_license')
        mock_click_secho.assert_called_with(
            'template',
            fg=self.github.config.clr_message)

    @mock.patch('gitsome.github.click.secho')
    def test_license_invalid(self, mock_click_secho):
        self.github.license('invalid_license')
        mock_click_secho.assert_called_with(
            ('  Invalid case-sensitive license requested, run the '
             'following command to see available licenses:\n'
             '    gh licenses'),
            fg=self.github.config.clr_error)

    @mock.patch('gitsome.github.click.secho')
    def test_licenses(self, mock_click_secho):
        self.github.licenses()
        mock_click_secho.assert_any_call(formatted_licenses)
        mock_click_secho.assert_any_call(formatted_licenses_tip,
                                         fg=self.github.config.clr_message)

    @mock.patch('gitsome.github.click.secho')
    def test_notifications(self, mock_click_secho):
        self.github.notifications()
        mock_click_secho.assert_called_with(formatted_threads)

    @mock.patch('gitsome.github.click.secho')
    def test_octocat(self, mock_click_secho):
        self.github.octocat('foo\\nbar')
        mock_click_secho.assert_called_with(
            'foo\nbar',
            fg=self.github.config.clr_message)

    @mock.patch('gitsome.github.click.secho')
    def test_pull_requests(self, mock_click_secho):
        self.github.pull_requests()
        mock_click_secho.assert_called_with(formatted_pull_requests)

    @mock.patch('gitsome.github.click.secho')
    def test_rate_limit(self, mock_click_secho):
        self.github.rate_limit()
        mock_click_secho.assert_called_with(
            'Rate limit: 5000',
            fg=self.github.config.clr_message)

    @mock.patch('gitsome.web_viewer.WebViewer.view_url')
    def test_repository(self, mock_view_url):
        self.github.repository('user1/repo1')
        mock_view_url.assert_called_with(
            'https://github.com/user1/repo1')

    @mock.patch('gitsome.github.click.secho')
    def test_repository_invalid(self, mock_click_secho):
        self.github.repository('user1/repo1/1')
        mock_click_secho.assert_called_with(
            'Expected argument: user/repo.',
            fg=self.github.config.clr_error)

    @mock.patch('gitsome.github.click.secho')
    @mock.patch('gitsome.github.GitHub.issues')
    def test_search_issues(self, mock_github_issues, mock_click_secho):
        self.github.search_issues('foo')
        mock_github_issues.assert_called_with(
            ['foobar', 'foobar', 'foobar'], 1000, False, sort=False)

    @mock.patch('gitsome.github.click.secho')
    @mock.patch('gitsome.github.GitHub.repositories')
    def test_search_repos(self, mock_github_repositories, mock_click_secho):
        self.github.search_repositories('foo', 'stars')
        mock_github_repositories.assert_called_with(
            ['foobar'], 1000, False, sort=False)

    @mock.patch('gitsome.github.click.secho')
    def test_trending(self, mock_click_secho):
        self.github.trending('Python', False, False, False)
        mock_click_secho.assert_called_with(formatted_trends)

    @mock.patch('gitsome.github.click.secho')
    def test_user(self, mock_click_secho):
        self.github.user('user1')
        mock_click_secho.assert_called_with(formatted_user)
        self.github.user('user2')
        mock_click_secho.assert_called_with(formatted_org)

    @mock.patch('gitsome.github.click.secho')
    def test_user_invalid(self, mock_click_secho):
        self.github.user('invalid_user')
        mock_click_secho.assert_called_with(
            'Invalid user.',
            fg=self.github.config.clr_error)

    @mock.patch('gitsome.github.click.secho')
    @mock.patch('gitsome.github.webbrowser.open')
    def test_user_browser(self, mock_webbrowser_open, mock_click_secho):
        self.github.user('invalid_user', browser=True)
        mock_webbrowser_open.assert_called_with(
            'https://github.com/invalid_user')

    @mock.patch('gitsome.github.click.secho')
    @mock.patch('gitsome.github.webbrowser.open')
    def test_view_browser(self, mock_webbrowser_open, mock_click_secho):
        self.github.config.load_urls = lambda x: ['user1/foo']
        self.github.view(1, view_in_browser=True)
        mock_webbrowser_open.assert_called_with(
            'https://github.com/user1/foo')

    @mock.patch('gitsome.github.click.secho')
    @mock.patch('gitsome.github.GitHub.issue')
    def test_view_issue(self, mock_github_issue, mock_click_secho):
        self.github.config.load_urls = lambda x: ['user1/foo/issues/1']
        self.github.view(0)
        mock_github_issue.assert_called_with('user1/foo/1')

    @mock.patch('gitsome.github.click.secho')
    @mock.patch('gitsome.github.GitHub.repository')
    def test_view_repo(self, mock_github_repository, mock_click_secho):
        self.github.config.load_urls = lambda x: ['user1/foo']
        self.github.view(0)
        mock_github_repository.assert_called_with('user1/foo')

    @mock.patch('gitsome.github.click.secho')
    @mock.patch('gitsome.web_viewer.WebViewer.view_url')
    def test_view_user(self, mock_view_url, mock_click_secho):
        self.github.config.load_urls = lambda x: ['user1']
        self.github.view(0)
        mock_view_url.assert_called_with('https://github.com/user1')

    def test_base_url(self):
        self.github.config.enterprise_url = 'https://github.intra.example.com'
        assert self.github.base_url == 'https://github.intra.example.com'
        self.github.config.enterprise_url = None
        assert self.github.base_url == self.github._base_url

    def test_add_base_url(self):
        expected = self.github.base_url + 'foo.html'
        assert self.github.add_base_url('foo.html') == expected
        assert self.github.add_base_url(expected) == expected



# -*- coding: utf-8 -*-
# Code from https://github.com/hit9/img2txt.git by Chao Wang
import sys
from docopt import docopt


def HTMLColorToRGB(colorstring):
    """ convert #RRGGBB to an (R, G, B) tuple """
    colorstring = colorstring.strip()
    if colorstring[0] == '#':
        colorstring = colorstring[1:]
    if len(colorstring) != 6:
        raise ValueError(
            "input #{0} is not in #RRGGBB format".format(colorstring))
    r, g, b = colorstring[:2], colorstring[2:4], colorstring[4:]
    r, g, b = [int(n, 16) for n in (r, g, b)]
    return (r, g, b)


def alpha_blend(src, dst):
    # Does not assume that dst is fully opaque
    # See https://en.wikipedia.org/wiki/Alpha_compositing - section on "Alpha
    # Blending"
    src_multiplier = (src[3] / 255.0)
    dst_multiplier = (dst[3] / 255.0) * (1 - src_multiplier)
    result_alpha = src_multiplier + dst_multiplier
    if result_alpha == 0:       # special case to prevent div by zero below
        return (0, 0, 0, 0)
    else:
        return (
            int(((src[0] * src_multiplier) +
                (dst[0] * dst_multiplier)) / result_alpha),
            int(((src[1] * src_multiplier) +
                (dst[1] * dst_multiplier)) / result_alpha),
            int(((src[2] * src_multiplier) +
                (dst[2] * dst_multiplier)) / result_alpha),
            int(result_alpha * 255)
        )


def getANSIcolor_for_rgb(rgb):
    # Convert to web-safe color since that's what terminals can handle in
    # "256 color mode"
    #   https://en.wikipedia.org/wiki/ANSI_escape_code
    # http://misc.flogisoft.com/bash/tip_colors_and_formatting#bash_tipscolors_and_formatting_ansivt100_control_sequences # noqa
    # http://superuser.com/questions/270214/how-can-i-change-the-colors-of-my-xterm-using-ansi-escape-sequences # noqa
    websafe_r = int(round((rgb[0] / 255.0) * 5))
    websafe_g = int(round((rgb[1] / 255.0) * 5))
    websafe_b = int(round((rgb[2] / 255.0) * 5))
    # Return ANSI coolor
    # https://en.wikipedia.org/wiki/ANSI_escape_code (see 256 color mode
    # section)
    return int(((websafe_r * 36) + (websafe_g * 6) + websafe_b) + 16)


def getANSIfgarray_for_ANSIcolor(ANSIcolor):
    """Return array of color codes to be used in composing an SGR escape
    sequence. Using array form lets us compose multiple color updates without
    putting out additional escapes"""
    # We are using "256 color mode" which is available in xterm but not
    # necessarily all terminals
    # To set FG in 256 color you use a code like ESC[38;5;###m
    return ['38', '5', str(ANSIcolor)]


def getANSIbgarray_for_ANSIcolor(ANSIcolor):
    """Return array of color codes to be used in composing an SGR escape
    sequence. Using array form lets us compose multiple color updates without
    putting out additional escapes"""
    # We are using "256 color mode" which is available in xterm but not
    # necessarily all terminals
    # To set BG in 256 color you use a code like ESC[48;5;###m
    return ['48', '5', str(ANSIcolor)]


def getANSIbgstring_for_ANSIcolor(ANSIcolor):
    # Get the array of color code info, prefix it with ESCAPE code and
    # terminate it with "m"
    return "\x1b[" + ";".join(getANSIbgarray_for_ANSIcolor(ANSIcolor)) + "m"


def generate_ANSI_to_set_fg_bg_colors(cur_fg_color, cur_bg_color, new_fg_color,
                                      new_bg_color):
    # This code assumes that ESC[49m and ESC[39m work for resetting bg and fg
    # This may not work on all terminals in which case we would have to use
    # ESC[0m
    # to reset both at once, and then put back fg or bg that we actually want

    # We don't change colors that are already the way we want them - saves
    # lots of file size

    # use array mechanism to avoid multiple escape sequences if we need to
    # change fg and bg
    color_array = []

    if new_bg_color != cur_bg_color:
        if new_bg_color is None:
            color_array.append('49')        # reset to default
        else:
            color_array += getANSIbgarray_for_ANSIcolor(new_bg_color)
    if new_fg_color != cur_fg_color:
        if new_fg_color is None:
            color_array.append('39')        # reset to default
        else:
            color_array += getANSIfgarray_for_ANSIcolor(new_fg_color)
    if len(color_array) > 0:
        return "\x1b[" + ";".join(color_array) + "m"
    else:
        return ""


def generate_ANSI_from_pixels(pixels, width, height, bgcolor_rgba,
                              get_pixel_func=None, is_overdraw=False):
    """Does not output final newline or reset to particular colors at end --
    caller should do that if desired bgcolor_rgba=None is treated as default
    background color."""
    if get_pixel_func is None:
        # just treat pixels as 2D array
        get_pixel_func = lambda pixels, x, y: (" ", pixels[x, y])
    # Compute ANSI bg color and strings we'll use to reset colors when moving
    # to next line
    if bgcolor_rgba is not None:
        bgcolor_ANSI = getANSIcolor_for_rgb(bgcolor_rgba)
        # Reset cur bg color to bgcolor because \n will fill the new line with
        # this color
        bgcolor_ANSI_string = getANSIbgstring_for_ANSIcolor(bgcolor_ANSI)
    else:
        bgcolor_ANSI = None
        # Reset cur bg color default because \n will fill the new line with
        # this color
        # reset bg to default (if we want to support terminals that can't
        # handle this will need to instead use 0m which clears fg too and then
        # when using this reset prior_fg_color to None too
        bgcolor_ANSI_string = "\x1b[49m"
    # removes all attributes (formatting and colors) to start in a known state
    string = "\x1b[0m"
    prior_fg_color = None       # this is an ANSI color not rgba
    prior_bg_color = None       # this is an ANSI color not rgba
    cursor_x = 0

    for h in range(height):
        for w in range(width):
            draw_char, rgba = get_pixel_func(pixels, w, h)
            # Handle fully or partially transparent pixels - but not if it is
            # the special "erase" character (None)
            skip_pixel = False
            if draw_char is not None:
                alpha = rgba[3]
                if alpha == 0:
                    skip_pixel = True       # skip any full transparent pixel
                elif alpha != 255 and bgcolor_rgba is not None:
                    # non-opaque so blend with specified bgcolor
                    rgba = alpha_blend(rgba, bgcolor_rgba)

            if not skip_pixel:
                this_pixel_str = ""
                # Throw away alpha channel - can still have non-fully-opaque
                # alpha value here if bgcolor was partially transparent or if
                # no bgcolor and not fully transparent. Could make argument to
                # use threshold to decide if throw away (e.g. >50% transparent)
                # vs. consider opaque (e.g. <50% transparent) but at least for
                # now we just throw it away
                rgb = rgba[:3]
                # If we've got the special "erase" character turn it into
                # outputting a space using the bgcolor which if None will
                # just be a reset to default bg which is what we want
                if draw_char is None:
                    draw_char = " "
                    color = bgcolor_ANSI
                else:
                    # Convert from RGB to ansi color, using closest color
                    color = getANSIcolor_for_rgb(rgb)
                    # Optimization - if we're drawing a space and the color is
                    # the same as a specified bg color then just skip this. We
                    # need to make this check here because the conversion to
                    # ANSI can cause colors that didn't match to now match
                    # We cannot do this optimization in overdraw mode because
                    # we cannot assume that the bg color is already drawn at
                    # this location
                    if not is_overdraw and (draw_char == " ") and \
                            (color == bgcolor_ANSI):
                        skip_pixel = True
                if not skip_pixel:
                    if len(draw_char) > 1:
                        raise ValueError(
                            "Not allowing multicharacter draw strings")
                    # If we are not at the cursor x location (happens if we
                    # skip pixels) output sequence to get there
                    # This is how we implement transparency - we don't draw
                    # spaces, we skip via cursor moves
                    if cursor_x < w:
                        # **SIZE - Note that when the bgcolor is specified
                        # (not None) and not overdrawing another drawing
                        # (as in an animation case) an optimization could be
                        # performed to draw spaces rather than output cursor
                        # advances. This would use less
                        # size when advancing less than 3 columns since the min
                        # escape sequence here is len 4. Not implementing this
                        # now
                        # code to advance N columns ahead
                        string += "\x1b[{0}C".format(w - cursor_x)
                        cursor_x = w
                    # Generate the ANSI sequences to set the colors the way we
                    # want them
                    if draw_char == " ":
                        # **SIZE - If we are willing to assume terminals that
                        # support ECH (Erase Character) as specified in here
                        # http://vt100.net/docs/vt220-rm/chapter4.html we could
                        # replace long runs of same-color spaces with single
                        # ECH codes. Seems like it is only correct to do this
                        # if BCE is supported though (
                        # http://superuser.com/questions/249898/how-can-i-prevent-os-x-terminal-app-from-overriding-vim-colours-on-a-remote-syst) # noqa
                        # else "erase" would draw the _default_ background
                        # color not the currently set background color
                        # We are supposed to output a space, so we're going to
                        # need to change the background color. No, we can't
                        # output an "upper ascii" character that fills the
                        # entire foreground - terminals don't display these the
                        # same way, if at all.
                        # Since we're outputting a space we can leave the prior
                        # fg color intact as it won't be used
                        string += generate_ANSI_to_set_fg_bg_colors(
                            prior_fg_color, prior_bg_color, prior_fg_color,
                            color)
                        prior_bg_color = color
                    else:
                        # We're supposed to output a non-space character, so
                        # we're going to need to change the foreground color
                        # and make sure the bg is set appropriately
                        string += generate_ANSI_to_set_fg_bg_colors(
                            prior_fg_color, prior_bg_color, color,
                            bgcolor_ANSI)
                        prior_fg_color = color
                        prior_bg_color = bgcolor_ANSI
                    # Actually output the character
                    string += draw_char
                    cursor_x = cursor_x + 1
        # Handle end of line - unless last line which is NOP because we don't
        # want to do anything to the _line after_ our drawing
        if (h + 1) != height:
            # Reset bg color so \n fills with it
            string += bgcolor_ANSI_string
            prior_bg_color = bgcolor_ANSI       # because it has been reset
            # Move to next line. If this establishes a new line in the terminal
            # then it fills the _newly established line_
            # to EOL with current bg color. However, if cursor had been moved
            # up and this just goes back down to an existing
            # line, no filling occurs
            string += "\n"
            cursor_x = 0
    return string


def generate_HTML_for_image(pixels, width, height):
    string = ""
    # first go through the height,  otherwise will rotate
    for h in range(height):
        for w in range(width):
            rgba = pixels[w, h]
            # TODO - could optimize output size by keeping span open until we
            # hit end of line or different color/alpha
            #   Could also output just rgb (not rgba) when fully opaque - if
            #   fully opaque is prevalent in an image
            #   those saved characters would add up
            string += ("<span style=\"color:rgba({0}, {1}, {2}, {3});\">"
                       "▇</span>").format(
                rgba[0], rgba[1], rgba[2], rgba[3] / 255.0)
        string += "\n"
    return string


def generate_grayscale_for_image(pixels, width, height, bgcolor):
    # grayscale
    color = "MNHQ$OC?7>!:-;. "
    string = ""
    # first go through the height,  otherwise will rotate
    for h in range(height):
        for w in range(width):
            rgba = pixels[w, h]
            # If partial transparency and we have a bgcolor, combine with bg
            # color
            if rgba[3] != 255 and bgcolor is not None:
                rgba = alpha_blend(rgba, bgcolor)
            # Throw away any alpha (either because bgcolor was partially
            # transparent or had no bg color)
            # Could make a case to choose character to draw based on alpha but
            # not going to do that now...
            rgb = rgba[:3]
            string += color[int(sum(rgb) / 3.0 / 256.0 * 16)]
        string += "\n"
    return string


def load_and_resize_image(imgname, antialias, maxLen):
    from PIL import Image
    img = Image.open(imgname)
    # force image to RGBA - deals with palettized images (e.g. gif) etc.
    if img.mode != 'RGBA':
        img = img.convert('RGBA')
    # resize up or down so that longer side of image is maxLen
    if maxLen is not None:
        native_width, native_height = img.size
        rate = float(maxLen) / max(native_width, native_height)
        width = int(rate * native_width) * 2
        height = int(rate * native_height)
        if native_width != width or native_height != height:
            img = img.resize((width, height), Image.ANTIALIAS
                             if antialias else Image.NEAREST)
    return img


def img2txt(imgname, maxLen=35, clr='', ansi=False,
            html=False, fontSize='', bgcolor='', antialias=True):
    try:
        maxLen = float(maxLen)
    except:
        maxLen = 35.0   # default maxlen: 30px
    try:
        fontSize = int(fontSize)
    except:
        fontSize = 7
    try:
        # add fully opaque alpha value (255)
        bgcolor = HTMLColorToRGB(bgcolor) + (255, )
    except:
        bgcolor = None
    try:
        from PIL import Image
        img = load_and_resize_image(imgname, antialias, maxLen)
    except IOError:
        return "File not found: " + imgname
    except ImportError:
        return 'PIL not found.'
    # get pixels
    pixel = img.load()
    width, height = img.size
    if ansi:
        # Since the "current line" was not established by us, it has been
        # filled with the current background color in the
        # terminal. We have no ability to read the current background color
        # so we want to refill the line with either
        # the specified bg color or if none specified, the default bg color.
        if bgcolor is not None:
            # Note that we are making the assumption that the viewing terminal
            # supports BCE (Background Color Erase) otherwise we're going to
            # get the default bg color regardless. If a terminal doesn't
            # support BCE you can output spaces but you'd need to know how many
            # to output (too many and you get linewrap)
            fill_string = getANSIbgstring_for_ANSIcolor(
                getANSIcolor_for_rgb(bgcolor))
        else:
            # reset bg to default (if we want to support terminals that can't
            # handle this will need to instead use 0m which clears fg too and
            # then when using this reset prior_fg_color to None too
            fill_string = "\x1b[49m"
        fill_string += "\x1b[K"          # does not move the cursor
        result = fill_string
        result += generate_ANSI_from_pixels(pixel, width, height, bgcolor)
        # Undo residual color changes, output newline because
        # generate_ANSI_from_pixels does not do so
        # removes all attributes (formatting and colors)
        result += "\x1b[0m\n"
        return result
    else:
        if clr:
            # TODO - should handle bgcolor - probably by setting it as BG on
            # the CSS for the pre
            string = generate_HTML_for_image(pixel, width, height)
        else:
            string = generate_grayscale_for_image(
                pixel, width, height, bgcolor)
        if html:
            # wrap with html
            template = """<!DOCTYPE HTML>
            <html>
            <head>
              <meta http-equiv="content-type" content="text/html; charset=utf-8" />
              <style type="text/css" media="all">
                pre {
                  white-space: pre-wrap;       /* css-3 */
                  white-space: -moz-pre-wrap;  /* Mozilla, since 1999 */
                  white-space: -pre-wrap;      /* Opera 4-6 */
                  white-space: -o-pre-wrap;    /* Opera 7 */
                  word-wrap: break-word;       /* Internet Explorer 5.5+ */
                  font-family: 'Menlo', 'Courier New', 'Consola';
                  line-height: 1.0;
                  font-size: %dpx;
                }
              </style>
            </head>
            <body>
              <pre>%s</pre>
            </body>
            </html>
            """
            return template % (fontSize, string)
        else:
            return string



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ..utils import timestamp_parameter
from ..models import BaseComment
from ..users import User


class IssueComment(BaseComment):
    """The :class:`IssueComment <IssueComment>` object. This structures and
    handles the comments on issues specifically.

    Two comment instances can be checked like so::

        c1 == c2
        c1 != c2

    And is equivalent to::

        c1.id == c2.id
        c1.id != c2.id

    See also: http://developer.github.com/v3/issues/comments/
    """
    def _update_attributes(self, comment):
        super(IssueComment, self)._update_attributes(comment)

        user = comment.get('user')
        #: :class:`User <github3.users.User>` who made the comment
        self.user = User(user, self) if user else None

        #: Issue url (not a template)
        self.issue_url = comment.get('issue_url')

        #: Html url (not a template)
        self.html_url = comment.get('html_url')

    def _repr(self):
        return '<Issue Comment [{0}]>'.format(self.user.login)


def issue_comment_params(sort, direction, since):
    params = {}

    if sort in ('created', 'updated'):
        params['sort'] = sort

    if direction in ('asc', 'desc'):
        params['direction'] = direction

    since = timestamp_parameter(since)
    if since:
        params['since'] = since

    return params



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import unicode_literals
from __future__ import print_function

import mock
from tests.compat import unittest

from gitsome.github import GitHub
from tests.data.markdown import formatted_markdown, raw_markdown, ssl_error


class WebViewerTest(unittest.TestCase):

    def setUp(self):
        self.github = GitHub()

    def test_format_markdown(self):
        result = self.github.web_viewer.format_markdown(raw_markdown)
        assert result == formatted_markdown

    @mock.patch('gitsome.github.click.echo_via_pager')
    def test_view_url(self, mock_click_echo_via_pager):
        url = 'https://www.github.com/donnemartin/gitsome'
        self.github.web_viewer.view_url(url)
        assert mock_click_echo_via_pager.mock_calls

    @unittest.skip('Skipping test_view_url_ssl_error')
    @mock.patch('gitsome.github.click.echo_via_pager')
    def test_view_url_ssl_error(self, mock_click_echo_via_pager):
        """Temp skipping this test due to a change [undocumented?] in the way
        the requests ssl error sample website is handled:
            http://docs.python-requests.org/en/master/user/advanced/#ssl-cert-verification  # NOQA
        See https://github.com/donnemartin/gitsome/pull/64 for more details.
        """
        url = 'https://requestb.in'
        self.github.web_viewer.view_url(url)
        mock_click_echo_via_pager.assert_called_with(ssl_error, None)



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

formatted_issues = u'\x1b[35m  1.   \x1b[0mtitle1 \x1b[0m@user2 \x1b[0m\x1b[35m(user1/repo1/issues/1)\x1b[0m\n\x1b[32m        State: open       \x1b[0m\x1b[36mComments: 1     \x1b[0m\x1b[33mAssignee: user1      \x1b[0m\n\x1b[35m  2.   \x1b[0mtitle2 \x1b[0m@user2 \x1b[0m\x1b[35m(user1/repo1/issues/2)\x1b[0m\n\x1b[32m        State: open       \x1b[0m\x1b[36mComments: 1     \x1b[0m\x1b[33mAssignee: user1      \x1b[0m\n\x1b[35m  3.   \x1b[0mtitle3 \x1b[0m@user2 \x1b[0m\x1b[35m(user1/repo1/issues/3)\x1b[0m\n\x1b[32m        State: open       \x1b[0m\x1b[36mComments: 1     \x1b[0m\x1b[33mAssignee: user1      \x1b[0m\n  View the page for \x1b[0m\x1b[35m1 through \x1b[0m\x1b[35m3\x1b[0m with the following command:\n\x1b[0m\x1b[35m    gh view [#] \x1b[0moptional: [-b/--browser] [--help]\n\x1b[0m\x1b[0m'
formatted_pull_requests = u'\x1b[35m  1.   \x1b[0mtitle1 \x1b[0m@user2 \x1b[0m\x1b[35m(user1/repo1/issues/1)\x1b[0m\n\x1b[32m        State: open       \x1b[0m\x1b[36mComments: 1     \x1b[0m\x1b[33mAssignee: user1      \x1b[0m\n\x1b[35m  2.   \x1b[0mtitle1 \x1b[0m@user2 \x1b[0m\x1b[35m(user1/repo1/issues/1)\x1b[0m\n\x1b[32m        State: open       \x1b[0m\x1b[36mComments: 1     \x1b[0m\x1b[33mAssignee: user1      \x1b[0m\n\x1b[35m  3.   \x1b[0mtitle1 \x1b[0m@user2 \x1b[0m\x1b[35m(user1/repo1/issues/1)\x1b[0m\n\x1b[32m        State: open       \x1b[0m\x1b[36mComments: 1     \x1b[0m\x1b[33mAssignee: user1      \x1b[0m\n  View the page for \x1b[0m\x1b[35m1 through \x1b[0m\x1b[35m3\x1b[0m with the following command:\n\x1b[0m\x1b[35m    gh view [#] \x1b[0moptional: [-b/--browser] [--help]\n\x1b[0m\x1b[0m'
